import o3seespy as o3  # for testing only
import pytest

def test_elastic_isotropic():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)


def test_elastic_isotropic_prebuild():
    mat = o3.nd_material.ElasticIsotropic(None, e_mod=1.0, nu=1.0, rho=0.0)
    osi = o3.OpenSeesInstance(ndm=2)
    mat.build(osi)


def test_elastic_orthotropic():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.ElasticOrthotropic(osi, ex=1.0, ey=1.0, ez=1.0, nu_xy=1.0, nu_yz=1.0, nu_zx=1.0, gxy=1.0, gyz=1.0, gzx=1.0, rho=0.0)


def test_j2plasticity():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.J2Plasticity(osi, k_mod=1.0, g_mod=1.0, sig0=1.0, sig_inf=1.0, delta=1.0, big_h=1.0)


def test_drucker_prager():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.DruckerPrager(osi, k_mod=1.0, g_mod=1.0, sigma_y=1.0, rho=1.0, rho_bar=1.0, kinf=1.0, ko=1.0, delta1=1.0, delta2=1.0, big_h=1.0, theta=1.0, density=1.0, atm_pressure=101e3)


@pytest.mark.skip()
def test_damage2p():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.Damage2p(osi, fcc=1.0, fct=1.0, e_mod=1.0, ni=1.0, gt=1.0, gc=1.0, rho_bar=1.0, big_h=1.0, theta=1.0, tangent=1.0)


def test_plane_stress():
    osi = o3.OpenSeesInstance(ndm=3)
    mat_3d = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
    o3.nd_material.PlaneStress(osi, mat3d=mat_3d)


def test_plane_strain():
    osi = o3.OpenSeesInstance(ndm=2)
    mat_3d = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
    o3.nd_material.PlaneStrain(osi, mat3d=mat_3d)


def test_multiaxial_cyclic_plasticity():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.MultiaxialCyclicPlasticity(osi, rho=1.0, k_mod=1.0, g_mod=1.0, su=1.0, ho=1.0, h=1.0, m=1.0, beta=1.0, k_coeff=1.0)


def test_bounding_cam_clay():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.BoundingCamClay(osi, mass_density=1.0, big_c=1.0, bulk_mod=1.0, ocr=1.0, mu_o=1.0, alpha=1.0, lamb=1.0, h=1.0, m=1.0)


def test_plate_fiber():
    osi = o3.OpenSeesInstance(ndm=3)
    mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
    o3.nd_material.PlateFiber(osi, three_d=mat)


def test_fsam():
    osi = o3.OpenSeesInstance(ndm=3)
    s_x = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    s_y = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    conc = o3.uniaxial_material.ConcreteCM(osi, fpcc=1.0, epcc=1.0, ec=1.0, rc=1.0, xcrn=1.0, ft=1.0, et=1.0, rt=1.0, xcrp=1.0,
                                    gap_close=0)
    o3.nd_material.FSAM(osi, rho=1.0, s_x=s_x, s_y=s_y, conc=conc, rou_x=1.0, rou_y=1.0, nu=1.0, alfadow=1.0)


def test_manzari_dafalias():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.ManzariDafalias(osi, g0=1.0, nu=1.0, e_init=1.0, mc=1.0, c=1.0, lambda_c=1.0, e0=1.0, ksi=1.0, p_atm=1.0, m=1.0, h0=1.0, ch=1.0, nb=1.0, a0=1.0, nd=1.0, z_max=1.0, cz=1.0, den=1.0)


def test_pm4sand():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.PM4Sand(osi, d_r=1.0, g_o=1.0, h_po=1.0, den=1.0, p_atm=1.0, h_o=1.0, e_max=1.0, e_min=1.0, n_b=1.0, n_d=1.0, a_do=1.0, z_max=1.0, c_z=1.0, c_e=1.0, phi_cv=1.0, nu=1.0, g_degr=1.0, c_dr=1.0, c_kaf=1.0, q_bolt=1.0, r_bolt=1.0, m_par=1.0, f_sed=1.0, p_sed=1.0)


@pytest.mark.skip()
def test_stress_density_model():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.StressDensityModel(osi, den=1.0, e_init=1.0, big_a=1.0, n=1.0, nu=1.0, a1=1.0, b1=1.0, a2=1.0, b2=1.0, a3=1.0, b3=1.0, fd=1.0, mu_0=1.0, mu_cyc=1.0, sc=1.0, big_m=1.0, p_atm=1.0)


def test_acoustic_medium():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.AcousticMedium(osi, k_mod=1.0, rho=1.0)




import o3seespy as o3  # for testing only


def test_constant():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.time_series.Constant(osi, factor=1.0)


def test_linear():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.time_series.Linear(osi, factor=1.0)


def test_trig():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.time_series.Trig(osi, t_start=1.0, t_end=1.0, period=1.0, factor=1.0, shift=0.0, zero_shift=0.0)


def test_triangle():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.time_series.Triangle(osi, t_start=1.0, t_end=1.0, period=1.0, factor=1.0, shift=0.0, zero_shift=0.0)


def test_rectangular():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.time_series.Rectangular(osi, t_start=1.0, t_end=1.0, factor=1.0)


def test_pulse():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.time_series.Pulse(osi, t_start=1.0, t_end=1.0, period=1.0, width=0.5, shift=0.0, factor=1.0, zero_shift=0.0)

#
# def test_path():
#     osi = o3.OpenSeesInstance(ndm=2)
#     values = [1.0, 1.0]
#     time = [1.0, 1.0]
#     o3.time_series.Path(osi, dt=0.0, values=values, time=time, filepath='', file_time='', factor=1.0, start_time=0.0, use_last="string", prepend_zero="string")




from o3seespy.command.nd_material.base_material import NDMaterialBase


class PlaneStressUserMaterial(NDMaterialBase):
    """
    The PlaneStressUserMaterial NDMaterial Class
    
    This command is used to create the multi-dimensional concrete material model that is based on the damage mechanism
    and smeared crack model.
    """
    op_type = 'PlaneStressUserMaterial'

    def __init__(self, osi, nstatevs, nprops, fc, ft, fcu, epsc0, epscu, epstu, stc):
        """
        Initial method for PlaneStressUserMaterial

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        nstatevs: None
            
        nprops: int
            Number of material properties (usually 7)
        fc: float
            Concrete compressive strength at 28 days (positive)
        ft: float
            Concrete tensile strength (positive)
        fcu: float
            Concrete crushing strength (negative)
        epsc0: float
            Concrete strain at maximum strength (negative)
        epscu: float
            Concrete strain at crushing strength (negative)
        epstu: float
            Ultimate tensile strain (positive)
        stc: float
            Shear retention factor

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.nd_material.PlaneStressUserMaterial(osi, nstatevs=1, nprops=1, fc=1.0, ft=1.0, fcu=1.0, epsc0=1.0, epscu=1.0, epstu=1.0, stc=1.0)
        """
        self.osi = osi
        self.nstatevs = nstatevs
        self.nprops = int(nprops)
        self.fc = float(fc)
        self.ft = float(ft)
        self.fcu = float(fcu)
        self.epsc0 = float(epsc0)
        self.epscu = float(epscu)
        self.epstu = float(epstu)
        self.stc = float(stc)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.nstatevs, self.nprops, self.fc, self.ft, self.fcu, self.epsc0, self.epscu, self.epstu, self.stc]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class PlateFromPlaneStress(NDMaterialBase):
    """
    The PlateFromPlaneStress NDMaterial Class
    
    This command is used to create the multi-dimensional concrete material model that is based on the damage mechanism
    and smeared crack model.
    """
    op_type = 'PlateFromPlaneStress'

    def __init__(self, osi, pre_def_mat, outof_plane_modulus):
        """
        Initial method for PlateFromPlaneStress

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        pre_def_mat: obj
            Integer object identifying planestress material
        outof_plane_modulus: float
            Shear modulus for out of plane stresses

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
        >>> o3.nd_material.PlateFromPlaneStress(osi, pre_def_mat=mat, outof_plane_modulus=1.0)
        """
        self.osi = osi
        self.pre_def_mat = pre_def_mat
        self.outof_plane_modulus = float(outof_plane_modulus)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.pre_def_mat.tag, self.outof_plane_modulus]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class PlateRebar(NDMaterialBase):
    """
    The PlateRebar NDMaterial Class
    
    This command is used to create the multi-dimensional reinforcement material.
    """
    op_type = 'PlateRebar'

    def __init__(self, osi, pre_def_mat, sita):
        """
        Initial method for PlateRebar

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        pre_def_mat: obj
            Integer object identifying uniaxial material
        sita: float
            Define the angle of reinforcement layer, 90 (longitudinal), 0 (tranverse)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> mat = o3.uniaxial_material.Elastic(osi, 1.0)
        >>> o3.nd_material.PlateRebar(osi, pre_def_mat=mat, sita=1.0)
        """
        self.osi = osi
        self.pre_def_mat = pre_def_mat
        self.sita = float(sita)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.pre_def_mat.tag, self.sita]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class PlasticDamageConcretePlaneStress(NDMaterialBase):
    """
    The PlasticDamageConcretePlaneStress NDMaterial Class
    
    No documentation is available yet. If you have the manual, please let me know.
    """
    op_type = 'PlasticDamageConcretePlaneStress'

    def __init__(self, osi, e_mod, nu, ft, fc, beta, ap, an, bn):
        """
        Initial method for PlasticDamageConcretePlaneStress

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        e_mod: None
            
        nu: None
            
        ft: None
            
        fc: None
            
        beta: None
            
        ap: None
            
        an: None
            
        bn: None
            
        """
        self.osi = osi
        self.e_mod = e_mod
        self.nu = nu
        self.ft = ft
        self.fc = fc
        self.beta = beta
        self.ap = ap
        self.an = an
        self.bn = bn
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.e_mod, self.nu, self.ft, self.fc, self.beta, self.ap, self.an, self.bn]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)



import o3seespy as o3  # for testing only
import pytest


def test_plane_stress_user_material():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.nd_material.PlaneStressUserMaterial(osi, nstatevs=1, nprops=1, fc=1.0, ft=1.0, fcu=1.0, epsc0=1.0, epscu=1.0, epstu=1.0, stc=1.0)


def test_plate_from_plane_stress():
    osi = o3.OpenSeesInstance(ndm=2)
    mat = o3.nd_material.ElasticIsotropic(osi, e_mod=1.0, nu=1.0, rho=0.0)
    o3.nd_material.PlateFromPlaneStress(osi, pre_def_mat=mat, outof_plane_modulus=1.0)


def test_plate_rebar():
    osi = o3.OpenSeesInstance(ndm=2)
    mat = o3.uniaxial_material.Elastic(osi, 1.0)
    o3.nd_material.PlateRebar(osi, pre_def_mat=mat, sita=1.0)




from o3seespy.command.element.base_element import ElementBase


class BeamColumnJoint(ElementBase):
    """
    The BeamColumnJoint Element Class
    
    This command is used to construct a two-dimensional beam-column-joint element object. The element may be used with
    both two-dimensional and three-dimensional structures; however, load is transferred only in the plane of the
    element.

    
    """
    op_type = 'beamColumnJoint'

    def __init__(self, osi, ele_nodes, mat1, mat2, mat3, mat4, mat5, mat6, mat7, mat8, mat9, mat10, mat11, mat12, mat13, ele_height_fac=1.0, ele_width_fac=1.0):
        """
        Initial method for BeamColumnJoint

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes
        mat1: obj
            Uniaxial material object for left bar-slip spring at node 1
        mat2: obj
            Uniaxial material object for right bar-slip spring at node 1
        mat3: obj
            Uniaxial material object for interface-shear spring at node 1
        mat4: obj
            Uniaxial material object for lower bar-slip spring at node 2
        mat5: obj
            Uniaxial material object for upper bar-slip spring at node 2
        mat6: obj
            Uniaxial material object for interface-shear spring at node 2
        mat7: obj
            Uniaxial material object for left bar-slip spring at node 3
        mat8: obj
            Uniaxial material object for right bar-slip spring at node 3
        mat9: obj
            Uniaxial material object for interface-shear spring at node 3
        mat10: obj
            Uniaxial material object for lower bar-slip spring at node 4
        mat11: obj
            Uniaxial material object for upper bar-slip spring at node 4
        mat12: obj
            Uniaxial material object for interface-shear spring at node 4
        mat13: obj
            Uniaxial material object for shear-panel
        ele_height_fac: float, optional
            Floating point value (as a ratio to the total height of the element) to be considered for determination of
            the distance in between the tension-compression couples (optional, default: 1.0)
        ele_width_fac: float, optional
            Floating point value (as a ratio to the total width of the element) to be considered for determination of
            the distance in between the tension-compression couples (optional, default: 1.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> mats = [o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None) for x in range(13)]
        >>> o3.element.BeamColumnJoint(osi, ele_nodes, *mats, ele_height_fac=1.0, ele_width_fac=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat1 = mat1
        self.mat2 = mat2
        self.mat3 = mat3
        self.mat4 = mat4
        self.mat5 = mat5
        self.mat6 = mat6
        self.mat7 = mat7
        self.mat8 = mat8
        self.mat9 = mat9
        self.mat10 = mat10
        self.mat11 = mat11
        self.mat12 = mat12
        self.mat13 = mat13
        self.ele_height_fac = float(ele_height_fac)
        self.ele_width_fac = float(ele_width_fac)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat1.tag, self.mat2.tag, self.mat3.tag, self.mat4.tag, self.mat5.tag, self.mat6.tag, self.mat7.tag, self.mat8.tag, self.mat9.tag, self.mat10.tag, self.mat11.tag, self.mat12.tag, self.mat13.tag, self.ele_height_fac, self.ele_width_fac]
        self.to_process(osi)


class ElasticTubularJoint(ElementBase):
    """
    The ElasticTubularJoint Element Class
    
    This command is used to construct an ElasticTubularJoint element object, which models joint flexibility of tubular
    joints in two dimensional analysis of any structure having tubular joints.

    
    """
    op_type = 'ElasticTubularJoint'

    def __init__(self, osi, ele_nodes, brace__diameter, brace__angle, big_e, chord__diameter, chord__thickness, chord__angle):
        """
        Initial method for ElasticTubularJoint

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of two element nodes
        brace__diameter: float
            Outer diameter of brace
        brace__angle: float
            Angle between brace and chord axis 0 < brace_angle < 90
        big_e: float
            Young's modulus
        chord__diameter: float
            Outer diameter of chord
        chord__thickness: float
            Thickness of chord
        chord__angle: float
            Angle between chord axis and global x-axis 0 < chord_angle < 180

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> o3.element.ElasticTubularJoint(osi, ele_nodes=ele_nodes, brace__diameter=1.0, brace__angle=1.0, big_e=1.0, chord__diameter=1.0, chord__thickness=1.0, chord__angle=1.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.brace__diameter = float(brace__diameter)
        self.brace__angle = float(brace__angle)
        self.big_e = float(big_e)
        self.chord__diameter = float(chord__diameter)
        self.chord__thickness = float(chord__thickness)
        self.chord__angle = float(chord__angle)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.brace__diameter, self.brace__angle, self.big_e, self.chord__diameter, self.chord__thickness, self.chord__angle]
        self.to_process(osi)


class Joint2D(ElementBase):
    """
    The Joint2D Element Class
    
    This command is used to construct a two-dimensional beam-column-joint element object. The two dimensional
    beam-column joint is idealized as a parallelogram shaped shear panel with adjacent elements connected to its
    mid-points. The midpoints of the parallelogram are referred to as external nodes. These nodes are the only
    analysis components that connect the joint element to the surrounding structure.

    
    """
    op_type = 'Joint2D'

    def __init__(self, osi, ele_nodes, mat1, mat2, mat3, mat4, mat_c, lrg_dsp, dmg, dmg1dmg2dmg3dmg4dmg_c=None):
        """
        Initial method for Joint2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of five element nodes = ``[nd1,nd2,nd3,nd4,ndc]``. ``ndc`` is the central node of beam-column joint.
            (the object ``ndc`` is used to generate the internal node, thus, the node should not exist in the domain or be used by
            any other node)
        mat1: int
            Uniaxial material object for interface rotational spring at node 1. use a zero object to indicate the case
            that a beam-column element is rigidly framed to the joint. 
        mat2: int
            Uniaxial material object for interface rotational spring at node 2. use a zero object to indicate the case
            that a beam-column element is rigidly framed to the joint. 
        mat3: int
            Uniaxial material object for interface rotational spring at node 3. use a zero object to indicate the case
            that a beam-column element is rigidly framed to the joint. 
        mat4: int
            Uniaxial material object for interface rotational spring at node 4. use a zero object to indicate the case
            that a beam-column element is rigidly framed to the joint. 
        mat_c: int
            Uniaxial material object for rotational spring of the central node that describes shear panel behavior
        lrg_dsp: obj
            An integer indicating the flag for considering large deformations: * ``0`` - for small deformations and
            constant geometry * ``1`` - for large deformations and time varying geometry * ``2`` - for large deformations
            ,time varying geometry and length correction
        dmg: obj
            Damage model object
        dmg1dmg2dmg3dmg4dmg_c: None, optional
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1], [0.5, 0.5]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> o3.element.Joint2D(osi, ele_nodes=ele_nodes, mat1=1, mat2=1, mat3=1, mat4=1, mat_c=1, lrg_dsp='', dmg='', dmg1dmg2dmg3dmg4dmg_c=1)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat1 = int(mat1)
        self.mat2 = int(mat2)
        self.mat3 = int(mat3)
        self.mat4 = int(mat4)
        self.mat_c = int(mat_c)
        self.lrg_dsp = lrg_dsp
        self.dmg = dmg
        self.dmg1dmg2dmg3dmg4dmg_c = dmg1dmg2dmg3dmg4dmg_c
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat1, self.mat2, self.mat3, self.mat4, self.mat_c, self.lrg_dsp.tag, self.dmg.tag]
        if getattr(self, 'dmg1dmg2dmg3dmg4dmg_c') is not None:
            self._parameters += ['-damage', self.dmg1dmg2dmg3dmg4dmg_c]
        self.to_process(osi)



import sys
from tests import conftest

sys.path.append(conftest.EXAMPLES_DIR)
import cantilever_w_hinge_under_force_then_disp_loading
import frame_building_dynamic_analysis
import readme_example
import simple_sdof
import site_response_analysis
import pytest


def test_cantilever_w_hinge_under_force_then_disp_loading():
    cantilever_w_hinge_under_force_then_disp_loading.run(use_pload=0)


def test_frame_building_dynamic_analysis():
    frame_building_dynamic_analysis.run_as_e2e()


def test_readme_example():
    readme_example.run(show=0)


def test_simple_sdof():
    simple_sdof.test_sdof(show=0)


def test_site_response_analysis():
    site_response_analysis.run(show=0)



import o3seespy as o3  # for testing only
import pytest


def test_parameter():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.parameter(osi, p_args=1)

@pytest.mark.skip
def test_add_to_parameter():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.add_to_parameter(osi, p_args=1)


@pytest.mark.skip
def test_update_parameter():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.update_parameter(osi, new_value=1.0)


@pytest.mark.skip
def test_set_parameter():
    osi = o3.OpenSeesInstance(ndm=2)
    eles = [1, 1]
    o3.senscmds.set_parameter(osi, new_value=1.0, eles=eles, start=1, end=1, args=1)


def test_get_param_tags():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.get_param_tags(osi)


@pytest.mark.skip
def test_get_param_value():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.get_param_value(osi)


@pytest.mark.skip
def test_compute_gradients():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.compute_gradients(osi)


@pytest.mark.skip
def test_sensitivity_algorithm():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.sensitivity_algorithm(osi)


@pytest.mark.skip
def test_get_sens_node_disp():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.get_sens_node_disp(osi, dof=1, param=[])


@pytest.mark.skip
def test_get_sens_node_vel():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.get_sens_node_vel(osi, dof=1, param=[])


@pytest.mark.skip
def test_get_sens_node_accel():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.get_sens_node_accel(osi, dof=1, param=[])


@pytest.mark.skip
def test_get_sens_lambda():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.get_sens_lambda(osi, param=[])


@pytest.mark.skip
def test_get_sens_section_force():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.get_sens_section_force(osi, sec_num=1, dof=1, param=[])


@pytest.mark.skip
def test_get_sens_node_pressure():
    osi = o3.OpenSeesInstance(ndm=2)
    o3.senscmds.get_sens_node_pressure(osi, param=[])




from o3seespy.command.element.base_element import ElementBase
from o3seespy.command.node import Node
from o3seespy.command.element.zero_length import ZeroLength
from o3seespy.command.element.beam_column import ElasticBeamColumn2D
from o3seespy.command.geom_transf import Linear2D as geom_trans_Linear2D
from o3seespy import cc
from o3seespy.command import common

class BeamOnNonlinearWinklerFoundation(object):
    def __init__(self, top_nodes, bot_nodes, sf_eles, fd_eles, sf_mats=None):
        self.top_nodes = top_nodes
        self.bot_nodes = bot_nodes
        self.sf_eles = sf_eles
        self.fd_eles = fd_eles
        self.sf_mats = sf_mats


def gen_shallow_foundation_bnwf(osi, bottom_node, top_node, sf_mats, pos, fd_area, fd_e_mod, fd_iz, r_flag: float =None, orient: list =None):
        """
        Generates nodes and elements for a shallow foundation beam on nonlinear foundation

        Alpha Status - Subject to changes and testing

        Note: Only supports 2D

        Parameters
        ----------
        osi
        bottom_node
        top_node
        sf_mats
        pos
        fd_area
        fd_e_mod
        fd_iz
        r_flag
        orient

        Returns
        -------

        """
        # TODO: sf_horz_mats - if 1 then apply to left node, if len(sf_mats) apply to each
        # TODO: sf_frn - soil-foundation friction, if float, then add each sf_mat inside a flat bearing.
        if not hasattr(sf_mats, '__len__'):  # TODO: only supports 2D
            sf_mats = [sf_mats] * len(pos)
        else:
            sf_mats = list(sf_mats)
        injected = 0
        pos = list(pos)
        if top_node.x not in pos:
            pos.append(top_node.x)
            pos.sort()
            injected = 1
        ind = pos.index(top_node.x)
        if injected:
            sf_mats.insert(ind, None)  # insert a null material since there should not be a soil spring here

        top_nodes = []
        bot_nodes = []
        sf_eles = []
        fd_eles = []
        transf = geom_trans_Linear2D(osi, [])
        x_centre = bottom_node.x
        y_centre = bottom_node.y
        for i, x in enumerate(pos):
            top_nodes.append(Node(osi, x_centre + x, y_centre))
            bot_nodes.append(Node(osi, x_centre + x, y_centre))
            common.Fix3DOF(osi, bot_nodes[i], cc.FIXED, cc.FIXED, cc.FIXED)
            if sf_mats[i] is not None:
                sf_eles.append(ZeroLength(osi, [top_nodes[i], bot_nodes[i]], [sf_mats[i]], [cc.DOF2D_Y], r_flag, orient))
            if i != 0:
                fd_eles.append(ElasticBeamColumn2D(osi, [top_nodes[i-1], top_nodes[i]], fd_area, fd_e_mod, fd_iz, transf))
        common.EqualDOF(osi, top_node, top_nodes[ind], dofs=[cc.DOF2D_X, cc.DOF2D_Y, cc.DOF2D_ROTZ])
        
        return BeamOnNonlinearWinklerFoundation(top_nodes, bot_nodes, sf_eles, fd_eles, sf_mats)



from o3seespy.command.element.base_element import ElementBase


class Quad(ElementBase):
    """
    The Quad Element Class
    
    This command is used to construct a FourNodeQuad element object which uses a bilinear isoparametric formulation.

  
     
    """
    op_type = 'quad'

    def __init__(self, osi, ele_nodes, thick, otype, mat, pressure=0.0, rho=0.0, b1=0.0, b2=0.0):
        """
        Initial method for Quad

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        thick: float
            Element thickness
        otype: str
            String representing material behavior. the type parameter can be either ``'planestrain'`` or
            ``'planestress'``
        mat: obj
            Object of ndmaterial
        pressure: float, optional
            Surface pressure (optional, default = 0.0)
        rho: float, optional
            Element mass density (per unit volume) from which a lumped element mass matrix is computed (optional,
            default=0.0)
        b1: float, optional
            Constant body forces defined in the isoparametric domain (optional, default=0.0)
        b2: float, optional
            Constant body forces defined in the isoparametric domain (optional, default=0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2, ndf=2)
        >>> obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> o3.element.Quad(osi, ele_nodes=ele_nodes, thick=1.0, otype='PlaneStrain', mat=obj, pressure=1.0, rho=1.0, b1=0.0, b2=0.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.thick = float(thick)
        self.otype = otype
        self.mat = mat
        self.pressure = float(pressure)
        self.rho = float(rho)
        self.b1 = float(b1)
        self.b2 = float(b2)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.thick, self.otype, self.mat.tag, self.pressure, self.rho, self.b1, self.b2]
        self.to_process(osi)


class ShellMITC4(ElementBase):
    """
    The ShellMITC4 Element Class
    
    This command is used to construct a ShellMITC4 element object, which uses a bilinear isoparametric formulation in
    combination with a modified shear interpolation to improve thin-plate bending performance.

    
    """
    op_type = 'ShellMITC4'

    def __init__(self, osi, ele_nodes, sec):
        """
        Initial method for ShellMITC4

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        sec: obj
            Object associated with previously-defined sectionforcedeformation object. currently must be either a
            ``'platefibersection'``, or ``'elasticmembraneplatesection'``

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ShellMITC4(osi, ele_nodes=ele_nodes, sec=sec)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        self.to_process(osi)


class ShellDKGQ(ElementBase):
    """
    The ShellDKGQ Element Class
    
    This command is used to construct a ShellDKGQ element object, which is a quadrilateral shell element based on the
    theory of generalized conforming element.

    
    """
    op_type = 'ShellDKGQ'

    def __init__(self, osi, ele_nodes, sec):
        """
        Initial method for ShellDKGQ

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        sec: obj
            Object associated with previously-defined sectionforcedeformation object. currently can be a
            ``'platefibersection'``, a ``'elasticmembraneplatesection'`` and a ``'layeredshell'`` section

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ShellDKGQ(osi, ele_nodes=ele_nodes, sec=sec)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        self.to_process(osi)


class ShellDKGT(ElementBase):
    """
    The ShellDKGT Element Class
    
    This command is used to construct a ShellDKGT element object, which is a triangular shell element based on the
    theory of generalized conforming element.

    
    """
    op_type = 'ShellDKGT'

    def __init__(self, osi, ele_nodes, sec):
        """
        Initial method for ShellDKGT

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of three element nodes in clockwise or counter-clockwise order
        sec: obj
            Object associated with previously-defined sectionforcedeformation object. currently can be a
            ``'platefibersection'``, a ``'elasticmembraneplatesection'`` and a ``'layeredshell'`` section

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ShellDKGT(osi, ele_nodes=ele_nodes, sec=sec)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        self.to_process(osi)


class ShellNLDKGQ(ElementBase):
    """
    The ShellNLDKGQ Element Class
    
    This command is used to construct a ShellNLDKGQ element object accounting for the geometric nonlinearity of large
    deformation using the updated Lagrangian formula, which is developed based on the ShellDKGQ element.

    
    """
    op_type = 'ShellNLDKGQ'

    def __init__(self, osi, ele_nodes, sec):
        """
        Initial method for ShellNLDKGQ

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        sec: obj
            Object associated with previously-defined sectionforcedeformation object. currently can be a
            ``'platefibersection'``, a ``'elasticmembraneplatesection'`` and a ``'layeredshell'`` section

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ShellNLDKGQ(osi, ele_nodes=ele_nodes, sec=sec)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        self.to_process(osi)


class ShellNLDKGT(ElementBase):
    """
    The ShellNLDKGT Element Class
    
    This command is used to construct a ShellNLDKGT element object accounting for the geometric nonlinearity of large
    deformation using the updated Lagrangian formula, which is developed based on the ShellDKGT element.

    
    """
    op_type = 'ShellNLDKGT'

    def __init__(self, osi, ele_nodes, sec):
        """
        Initial method for ShellNLDKGT

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of three element nodes in clockwise or counter-clockwise order around the element
        sec: obj
            Object associated with previously-defined sectionforcedeformation object. currently can be a
            ``'platefibersection'``, a ``'elasticmembraneplatesection'`` and a ``'layeredshell'`` section

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ShellNLDKGT(osi, ele_nodes=ele_nodes, sec=sec)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        self.to_process(osi)


class ShellNL(ElementBase):
    """
    The ShellNL Element Class
    
    
    """
    op_type = 'ShellNL'

    def __init__(self, osi, ele_nodes, sec):
        """
        Initial method for ShellNL

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of nine element nodes, input is the typical, firstly four corner nodes counter-clockwise, then
            mid-side nodes counter-clockwise and finally the central node
        sec: obj
            Object associated with previously-defined sectionforcedeformation object. currently can be a
            ``'platefibersection'``, a ``'elasticmembraneplatesection'`` and a ``'layeredshell'`` section

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1], [0.5, 0], [1, 0.5], [0.5, 1], [0, 0.5], [0.5, 0.5]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
        >>> sec = o3.section.Elastic2D(osi, 10.0, 1.0, 1.0)
        >>> o3.element.ShellNL(osi, ele_nodes=ele_nodes, sec=sec)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.sec = sec
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.sec.tag]
        self.to_process(osi)


class BbarQuad(ElementBase):
    """
    The BbarQuad Element Class
    
    This command is used to construct a four-node quadrilateral element object, which uses a bilinear isoparametric
    formulation along with a mixed volume/pressure B-bar assumption. This element is for plane strain problems only.

    
    """
    op_type = 'bbarQuad'

    def __init__(self, osi, ele_nodes, thick, mat):
        """
        Initial method for BbarQuad

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        thick: float
            Element thickness
        mat: obj
            Object of ndmaterial

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2, ndf=2)
        >>> mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> o3.element.BbarQuad(osi, ele_nodes=ele_nodes, thick=1.0, mat=mat)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.thick = float(thick)
        self.mat = mat
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.thick, self.mat.tag]
        self.to_process(osi)


class EnhancedQuad(ElementBase):
    """
    The EnhancedQuad Element Class
    
    This command is used to construct a four-node quadrilateral element, which uses a bilinear isoparametric formulation
    with enhanced strain modes.

    
    """
    op_type = 'enhancedQuad'

    def __init__(self, osi, ele_nodes, thick, otype, mat):
        """
        Initial method for EnhancedQuad

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        thick: float
            Element thickness
        otype: str
            String representing material behavior. valid options depend on the ndmaterial object and its available
            material formulations. the type parameter can be either ``'planestrain'`` or ``'planestress'``
        mat: obj
            Object of ndmaterial

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2, ndf=2)
        >>> obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> o3.element.EnhancedQuad(osi, ele_nodes=ele_nodes, thick=1.0, otype='PlaneStress', mat=obj)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.thick = float(thick)
        self.otype = otype
        self.mat = mat
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.thick, self.otype, self.mat.tag]
        self.to_process(osi)


class SSPquad(ElementBase):
    """
    The SSPquad Element Class
    
    This command is used to construct a SSPquad element object.

    
    """
    op_type = 'SSPquad'

    def __init__(self, osi, ele_nodes, mat, otype, thick, b1, b2):
        """
        Initial method for SSPquad

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ele_nodes: list
            A list of four element nodes in counter-clockwise order
        mat: obj
            Unique integer object associated with previously-defined ndmaterial object
        otype: str
            String to relay material behavior to the element, can be either ``'planestrain'`` or ``'planestress'``
        thick: float
            Thickness of the element in out-of-plane direction
        b1: float
            Constant body forces in global x- and y-directions, respectively (optional, default = 0.0)
        b2: float
            Constant body forces in global x- and y-directions, respectively (optional, default = 0.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> obj = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
        >>> coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
        >>> ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(4)]
        >>> o3.element.SSPquad(osi, ele_nodes=ele_nodes, mat=obj, otype='PlaneStrain', thick=1.0, b1=0.0, b2=0.0)
        """
        self.osi = osi
        self.ele_node_tags = [x.tag for x in ele_nodes]
        self.ele_nodes = ele_nodes
        self.mat = mat
        self.otype = otype
        self.thick = float(thick)
        self.b1 = float(b1)
        self.b2 = float(b2)
        osi.n_ele += 1
        self._tag = osi.n_ele
        self._parameters = [self.op_type, self._tag, *self.ele_node_tags, self.mat.tag, self.otype, self.thick, self.b1, self.b2]
        self.to_process(osi)



import o3seespy as o3  # for testing only
import pytest


def test_elastomeric_bearing_plasticity2d():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    o3.element.ElastomericBearingPlasticity2D(osi, ele_nodes=ele_nodes, k_init=1.0, qd=1.0, alpha1=1.0, alpha2=1.0,
                                              mu=1.0, p_mat=p_mat, mz_mat=mz_mat)


def test_elastomeric_bearing_plasticity3d():
    osi = o3.OpenSeesInstance(ndm=3, ndf=6)
    coords = [[0, 0, 0], [0, 1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    t_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    my_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    orient_vals = [1, 0, 0]
    o3.element.ElastomericBearingPlasticity3D(osi, ele_nodes=ele_nodes, k_init=1.0, qd=1.0, alpha1=1.0, alpha2=1.0,
                                              mu=1.0, p_mat=p_mat, t_mat=t_mat, my_mat=my_mat, mz_mat=mz_mat,
                                              orient=orient_vals)


def test_elastomeric_bearing_bouc_wen2d():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    orient_vals = [1, 0, 0, 1, 0, 1]
    p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    o3.element.ElastomericBearingBoucWen2D(osi, ele_nodes=ele_nodes, k_init=1.0, qd=1.0, alpha1=1.0, alpha2=1.0,
                                           mu=1.0, eta=1.0, beta=1.0, gamma=1.0, p_mat=p_mat, mz_mat=mz_mat,
                                           orient_vals=orient_vals, shear_dist=1.0, do_rayleigh=False, mass=1.0)


def test_elastomeric_bearing_bouc_wen3d():
    osi = o3.OpenSeesInstance(ndm=3, ndf=6)
    coords = [[0, 0, 0], [0, 1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    orient_vals = [1, 0, 0]
    p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    t_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    my_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    o3.element.ElastomericBearingBoucWen3D(osi, ele_nodes=ele_nodes, k_init=1.0, qd=1.0, alpha1=1.0, alpha2=1.0,
                                           mu=1.0, eta=1.0, beta=1.0, gamma=1.0, p_mat=p_mat, t_mat=t_mat,
                                           my_mat=my_mat, mz_mat=mz_mat, orient_vals=orient_vals,
                                           shear_dist=1.0, do_rayleigh=False, mass=1.0)


def test_flat_slider_bearing2d():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    frn1 = o3.friction_model.Coulomb(osi, mu=1.0)
    o3.element.FlatSliderBearing2D(osi, ele_nodes=ele_nodes, frn_mdl=frn1, k_init=1.0, p_mat=p_mat, mz_mat=mz_mat,
                                   do_rayleigh=False, max_iter=1, tol=1.0, orient=None, mass=1.0, shear_dist=1.0)


def test_flat_slider_bearing3d():
    osi = o3.OpenSeesInstance(ndm=3, ndf=6)
    coords = [[0, 0, 0], [0, 1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    t_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    my_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    frn1 = o3.friction_model.Coulomb(osi, mu=1.0)
    orient_vals = [1, 0, 0]
    o3.element.FlatSliderBearing3D(osi, ele_nodes=ele_nodes, frn_mdl=frn1, k_init=1.0, p_mat=p_mat, t_mat=t_mat,
                                   my_mat=my_mat, mz_mat=mz_mat, do_rayleigh=False, max_iter=None, tol=None,
                                   mass=1.0, shear_dist=1.0, orient=orient_vals)


def test_single_fp_bearing2d():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    frn1 = o3.friction_model.Coulomb(osi, mu=1.0)
    o3.element.SingleFPBearing2D(osi, ele_nodes=ele_nodes, frn_mdl=frn1, reff=1.0, k_init=1.0, p_mat=p_mat,
                                 mz_mat=mz_mat, do_rayleigh=False, max_iter=1, tol=1.0, orient=None,
                                 mass=1.0, shear_dist=1.0)


def test_single_fp_bearing3d():
    osi = o3.OpenSeesInstance(ndm=3, ndf=6)
    coords = [[0, 0, 0], [0, 1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    p_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    mz_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    t_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    my_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    frn1 = o3.friction_model.Coulomb(osi, mu=1.0)
    orient_vals = [1, 0, 0]
    o3.element.SingleFPBearing3D(osi, ele_nodes=ele_nodes, frn_mdl=frn1, reff=1.0, k_init=1.0, p_mat=p_mat, t_mat=t_mat,
                                 my_mat=my_mat, mz_mat=mz_mat, do_rayleigh=False, max_iter=None, tol=None,
                                 orient=orient_vals, mass=1.0, shear_dist=1.0)


def test_tfp():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    o3.element.TFP(osi, ele_nodes=ele_nodes,
                   r1=1.0, r2=1.0, r3=1.0, r4=1.0,
                   db1=1.0, db2=1.0, db3=1.0, db4=1.0,
                   d1=1.0, d2=1.0, d3=1.0, d4=1.0,
                   mu1=0.3, mu2=0.4, mu3=0.5, mu4=0.5,
                   h1=1.0, h2=1.0, h3=1.0, h4=1.0,
                   h0=1.0, col_load=1.0, big_k=None)


@pytest.mark.skip()
def test_triple_friction_pendulum():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    frn1 = o3.friction_model.Coulomb(osi, mu=1.0)
    frn2 = o3.friction_model.Coulomb(osi, mu=1.0)
    frn3 = o3.friction_model.Coulomb(osi, mu=1.0)
    vert_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    rot_z_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    rot_x_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    rot_y_mat = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    o3.element.TripleFrictionPendulum(osi, ele_nodes=ele_nodes, frn1=frn1, frn2=frn2, frn3=frn3, vert_mat=vert_mat,
                                      rot_z_mat=rot_z_mat, rot_x_mat=rot_x_mat, rot_y_mat=rot_y_mat, l1=1.0, l2=1.0,
                                      l3=1.0, d1=1.0, d2=1.0, d3=1.0, big_w=1.0, uy=1.0, kvt=1.0, min_fv=None, tol=1.0)


def test_multiple_shear_spring():
    osi = o3.OpenSeesInstance(ndm=3, ndf=6)
    coords = [[0, 0, 0], [1, 0, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat = o3.uniaxial_material.Elastic(osi, 1.0)
    o3.element.MultipleShearSpring(osi, ele_nodes=ele_nodes, n_spring=1, mat=mat, lim=1.0, mass=1.0, orient=None)


@pytest.mark.skip()
def test_kikuchi_bearingadjust_pd_output():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat_mss = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    mat_mns = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    o3.element.KikuchiBearingadjustPDOutput(osi, ele_nodes=ele_nodes, shape=1.0, size=1.0, total_rubber=1.0, total_height=1.0, n_mss=1, mat_mss=mat_mss, lim_disp=1.0, n_mns=1, mat_mns=mat_mns, lamb=1.0, no_pd_input="string", no_tilt="string", ci=1.0, cj=1.0, orient=[0.0, 0.0], mass=1.0)


@pytest.mark.skip()
def test_kikuchi_bearingdo_balance():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat_mss = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    mat_mns = o3.uniaxial_material.Elastic(osi, e_mod=1.0, eta=0.0, eneg=None)
    o3.element.KikuchiBearingdoBalance(osi, ele_nodes=ele_nodes, shape=1.0, size=1.0, total_rubber=1.0, total_height=1.0, n_mss=1, mat_mss=mat_mss, lim_disp=1.0, n_mns=1, mat_mns=mat_mns, lamb=1.0, no_pd_input="string", no_tilt="string", lim_fo=1.0, lim_fi=1.0, n_iter=1.0, orient=[0.0, 0.0], mass=1.0)


@pytest.mark.skip()
def test_yamamoto_biaxial_hd_rco_rs():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    o3.element.YamamotoBiaxialHDRcoRS(osi, ele_nodes=ele_nodes, tp=1, d_do=1.0, d_di=1.0, hr=1.0, cr=1.0, cs=1.0, orient=[0.0, 0.0], mass=1.0)


@pytest.mark.skip()
def test_elastomeric_x():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    o3.element.ElastomericX(osi, ele_nodes=ele_nodes, fy=1.0, alpha=1.0, gr=1.0, kbulk=1.0, d1=1.0, d2=1.0, ts=1.0,
                            tr=1.0, n=1, x1=1.0, x2=1.0, x3=1.0, y1=1.0, y2=1.0, y3=1.0, kc=1.0, phi_m=1.0, ac=1.0,
                            s_dratio=1.0, m=1.0, cd=1.0, tc=1.0, tag1=0, tag2=0, tag3=0, tag4=0)


def test_rj_watson_eqs_bearing2d():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    p_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
    vy_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
    mz_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
    frn_mdl = o3.friction_model.Coulomb(osi, mu=1.0)
    o3.element.RJWatsonEqsBearing2D(osi, ele_nodes=ele_nodes, frn_mdl=frn_mdl, k_init=1.0, p_mat=p_mat, vy_mat=vy_mat,
                                    mz_mat=mz_mat, do_rayleigh=False, max_iter=1, tol=1.0, orient=None, mass=1.0,
                                    shear_dist=1.0)


def test_rj_watson_eqs_bearing3d():
    osi = o3.OpenSeesInstance(ndm=3, ndf=6)
    coords = [[0, 0, 0], [0, 1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
    p_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
    vy_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
    vz_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
    t_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
    my_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
    mz_mat = o3.uniaxial_material.Elastic(osi, 1, 1)
    orient_vals = [1, 0, 0]
    frn_mdl = o3.friction_model.Coulomb(osi, mu=1.0)
    o3.element.RJWatsonEqsBearing3D(osi, ele_nodes=ele_nodes, frn_mdl=frn_mdl, k_init=1.0, p_mat=p_mat,
                                    vy_mat=vy_mat, vz_mat=vz_mat, t_mat=t_mat, my_mat=my_mat, mz_mat=mz_mat,
                                    do_rayleigh=False, max_iter=1, tol=1.0, orient=orient_vals, mass=1.0, shear_dist=1.0)


@pytest.mark.skip()
def test_lead_rubber_x():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    o3.element.LeadRubberX(osi, ele_nodes=ele_nodes, fy=1.0, alpha=1.0, gr=1.0, kbulk=1.0, d1=1.0, d2=1.0, ts=1.0,
                           tr=1.0, n=1, x1=1.0, x2=1.0, x3=1.0, y1=1.0, y2=1.0, y3=1.0, kc=1.0, phi_m=1.0, ac=1.0,
                           s_dratio=1.0, m=1.0, cd=1.0, tc=1.0, q_l=1.0, c_l=1.0, k_s=1.0, a_s=1.0,
                           tag1=1, tag2=1, tag3=1, tag4=1, tag5=1)


@pytest.mark.skip()
def test_hdr():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    o3.element.HDR(osi, ele_nodes=ele_nodes, gr=1.0, kbulk=1.0, d1=1.0, d2=1.0, ts=1.0, tr=1.0, n=1, a1=1.0, a2=1.0, a3=1.0, b1=1.0, b2=1.0, b3=1.0, c1=1.0, c2=1.0, c3=1.0, c4=1.0, x1=1.0, x2=1.0, x3=1.0, y1=1.0, y2=1.0, y3=1.0, kc=1.0, phi_m=1.0, ac=1.0, s_dratio=1.0, m=1.0, tc=1.0)


@pytest.mark.skip()
def test_fp_bearing_ptv():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    o3.element.FPBearingPTV(osi, ele_nodes=ele_nodes, mu_ref=1.0, is_pressure_dependent=1, p_ref=1.0, is_temperature_dependent=1, diffusivity=1.0, conductivity=1.0, is_velocity_dependent=1, rate_parameter=1.0, reffective_fp=1.0, radius__contact=1.0, k_initial=1.0, the_material_a=1, the_material_b=1, the_material_c=1, the_material_d=1, x1=1.0, x2=1.0, x3=1.0, y1=1.0, y2=1.0, y3=1.0, shear_dist=1.0, do_rayleigh=1, mass=1.0, max_iter=1, tol=1.0, unit=1)


if __name__ == '__main__':
    test_elastomeric_bearing_bouc_wen2d()



from o3seespy.base_model import OpenSeesObject


class PatternBase(OpenSeesObject):
    op_base_type = "pattern"


class Plain(PatternBase):
    """
    The Plain Pattern Class
    
    This commnand allows the user to construct a LoadPattern object. Each plain load pattern is associated with a
    TimeSeries object and can contain multiple NodalLoads, ElementalLoads and SP_Constraint objects. The command to
    generate LoadPattern object contains in { } the commands to generate all the loads and the single-point
    constraints in the pattern. To construct a load pattern and populate it, the following command is used:
    """
    op_type = 'Plain'

    def __init__(self, osi, ts, fact: float=None):
        """
        Initial method for Plain

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        ts: obj
            The object of the time series to be used in the load pattern
        fact: float, optional
            Constant factor. 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> ts = o3.time_series.Linear(osi, factor=1.0)
        >>> o3.pattern.Plain(osi, ts=ts, fact=1.0)
        """
        self.osi = osi
        self.ts = ts
        if fact is None:
            self.fact = None
        else:
            self.fact = float(fact)
        osi.n_pat += 1
        self._tag = osi.n_pat
        self._parameters = [self.op_type, self._tag, self.ts.tag]
        if getattr(self, 'fact') is not None:
            self._parameters += ['-fact', self.fact]
        self.to_process(osi)


class UniformExcitation(PatternBase):
    """
    The UniformExcitation Pattern Class
    
    The UniformExcitation pattern allows the user to apply a uniform excitation to a model acting in a certain
    direction. The command is as follows:
    """
    op_type = 'UniformExcitation'

    def __init__(self, osi, dir, disp_series=None, vel_series=None, accel_series=None, vel0: float=None, fact: float=None):
        """
        Initial method for UniformExcitation

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        dir: int
            Direction in which ground motion acts #. corresponds to translation along the global x axis #. corresponds
            to translation along the global y axis #. corresponds to translation along the global z axis #. corresponds to
            rotation about the global x axis #. corresponds to rotation about the global y axis #. corresponds to
            rotation about the global z axis
        disp_series: obj, optional
            Object of the timeseries series defining the displacement history. 
        vel_series: obj, optional
            Object of the timeseries series defining the velocity history. 
        accel_series: obj, optional
            Object of the timeseries series defining the acceleration history. 
        vel0: float, optional
            The initial velocity (optional, default=0.0)
        fact: float, optional
            Constant factor (optional, default=1.0)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> ts = o3.time_series.Linear(osi, factor=1.0)
        >>> o3.pattern.UniformExcitation(osi, dir=1, accel_series=ts, vel0=1.0, fact=1.0)
        """
        self.osi = osi
        self.dir = int(dir)
        self.disp_series = disp_series
        self.vel_series = vel_series
        self.accel_series = accel_series
        if vel0 is None:
            self.vel0 = None
        else:
            self.vel0 = float(vel0)
        if fact is None:
            self.fact = None
        else:
            self.fact = float(fact)
        osi.n_pat += 1
        self._tag = osi.n_pat
        self._parameters = [self.op_type, self._tag, self.dir]
        if getattr(self, 'disp_series') is not None:
            self._parameters += ['-disp', self.disp_series.tag]
        if getattr(self, 'vel_series') is not None:
            self._parameters += ['-vel', self.vel_series.tag]
        if getattr(self, 'accel_series') is not None:
            self._parameters += ['-accel', self.accel_series.tag]
        if getattr(self, 'vel0') is not None:
            self._parameters += ['-vel0', self.vel0]
        if getattr(self, 'fact') is not None:
            self._parameters += ['-fact', self.fact]
        self.to_process(osi)


class MultipleSupport(PatternBase):
    """
    The MultipleSupport Pattern Class
    
    The Multi-Support pattern allows similar or different prescribed ground motions to be input at various supports in
    the structure. In OpenSees, the prescribed motion is applied using single-point constraints, the single-point
    constraints taking their constraint value from user created ground motions.
    """
    op_type = 'MultipleSupport'

    def __init__(self, osi):
        """
        Initial method for MultipleSupport

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.pattern.MultipleSupport(osi)
        """
        self.osi = osi
        osi.n_pat += 1
        self._tag = osi.n_pat
        self._parameters = [self.op_type, self._tag]
        self.to_process(osi)




def random_variable(osi, dist, mean: float=None, stdv: float=None, start_point: float=None, params: list=None):
    """
    Create a random variable with user specified distribution

    Parameters
    ----------
    osi: o3seespy.OpenSeesInstance
    dist: str
            Random variable distribution * ``'normal'`` * ``'lognormal'`` * ``'gamma'`` * ``'shiftedexponential'`` *
            ``'shiftedrayleigh'`` * ``'exponential'`` * ``'rayleigh'`` * ``'uniform'`` * ``'beta'`` * ``'type1largestvalue'`` *
            ``'type1smallestvalue'`` * ``'type2largestvalue'`` * ``'type3smallestvalue'`` * ``'chisquare'`` * ``'gumbel'`` *
            ``'weibull'`` * ``'laplace'`` * ``'pareto'``
    mean: float, optional
            Mean value
    stdv: float, optional
            Standard deviation
    start_point: float, optional
            Starting point of the distribution
    params: list, optional
            A list of parameter objects
    """
    if mean is not None:
        mean = float(mean)
    if stdv is not None:
        stdv = float(stdv)
    if start_point is not None:
        start_point = float(start_point)
    _parameters = [dist]
    if mean is not None:
        _parameters += ['-mean', mean]
    if stdv is not None:
        _parameters += ['-stdv', stdv]
    if start_point is not None:
        _parameters += ['-startPoint', start_point]
    if params is not None:
        _parameters += ['-parameters', *params]
    return osi.to_process("randomVariable", _parameters)



import eqsig
from eqsig import sdof
import numpy as np

import o3seespy as o3


def get_inelastic_response(mass, k_spring, f_yield, motion, dt, xi=0.05, r_post=0.0):
    """
    Run seismic analysis of a nonlinear SDOF

    :param mass: SDOF mass
    :param k_spring: spring stiffness
    :param f_yield: yield strength
    :param motion: list, acceleration values
    :param dt: float, time step of acceleration values
    :param xi: damping ratio
    :param r_post: post-yield stiffness
    :return:
    """
    osi = o3.OpenSeesInstance(ndm=2)

    # Establish nodes
    bot_node = o3.node.Node(osi, 0, 0)
    top_node = o3.node.Node(osi, 0, 0)

    # Fix bottom node
    o3.Fix3DOF(osi, top_node, o3.cc.FREE, o3.cc.FIXED, o3.cc.FIXED)
    o3.Fix3DOF(osi, bot_node, o3.cc.FIXED, o3.cc.FIXED, o3.cc.FIXED)
    # Set out-of-plane DOFs to be slaved
    o3.EqualDOF(osi, top_node, bot_node, [o3.cc.Y, o3.cc.DOF2D_ROTZ])

    # nodal mass (weight / g):
    o3.Mass(osi, top_node, mass, 0., 0.)

    # Define material
    bilinear_mat = o3.uniaxial_material.Steel01(osi, fy=f_yield, e0=k_spring, b=r_post)

    # Assign zero length element, # Note: pass actual node and material objects into element
    o3.element.ZeroLength(osi, [bot_node, top_node], mats=[bilinear_mat], dirs=[o3.cc.DOF2D_X], r_flag=1)

    # Define the dynamic analysis
    values = list(-1 * motion)  # should be negative
    acc_series = o3.time_series.Path(osi, dt, values)
    o3.pattern.UniformExcitation(osi, o3.cc.X, accel_series=acc_series)

    # set damping based on first eigen mode
    angular_freq2 = o3.get_eigen(osi, solver='fullGenLapack', n=1)
    if hasattr(angular_freq2, '__len__'):
        angular_freq2 = angular_freq2[0]
    angular_freq = angular_freq2 ** 0.5
    beta_k = 2 * xi / angular_freq
    o3.rayleigh.Rayleigh(osi, alpha_m=0.0, beta_k=beta_k, beta_k_init=0.0, beta_k_comm=0.0)

    # Run the dynamic analysis

    o3.wipe_analysis(osi)
    newmark_gamma = 0.5
    newmark_beta = 0.25

    o3.algorithm.Newton(osi)
    o3.constraints.Transformation(osi)
    o3.algorithm.Newton(osi)
    o3.numberer.RCM(osi)
    o3.system.SparseGeneral(osi)
    o3.integrator.Newmark(osi, newmark_gamma, newmark_beta)
    o3.analysis.Transient(osi)

    o3.test_check.EnergyIncr(osi, tol=1.0e-10, max_iter=10)
    analysis_time = (len(values) - 1) * dt
    analysis_dt = 0.001
    outputs = {
        "time": [],
        "rel_disp": [],
        "rel_accel": [],
        "force": []
    }
    o3.record(osi)
    curr_time = o3.get_time(osi)
    while curr_time < analysis_time:
        outputs["time"].append(curr_time)
        outputs["rel_disp"].append(o3.get_node_disp(osi, top_node, o3.cc.X))
        outputs["rel_accel"].append(o3.get_node_accel(osi, top_node, o3.cc.X))
        o3.gen_reactions(osi)
        outputs["force"].append(-o3.get_node_reaction(osi, bot_node, o3.cc.X))  # Negative since diff node
        o3.analyze(osi, 1, analysis_dt)
        curr_time = o3.get_time(osi)
    o3.wipe(osi)
    for item in outputs:
        outputs[item] = np.array(outputs[item])

    return outputs


def test_sdof():
    """
    Create a plot of an elastic analysis, nonlinear analysis and closed form elastic

    :return:
    """

    from tests.conftest import TEST_DATA_DIR

    record_path = TEST_DATA_DIR
    record_filename = 'test_motion_dt0p01.txt'
    dt = 0.01
    rec = np.loadtxt(record_path + record_filename)
    acc_signal = eqsig.AccSignal(rec, dt)
    period = 1.0
    xi = 0.05
    mass = 1.0
    f_yield = 1.5  # Reduce this to make it nonlinear
    r_post = 0.0

    periods = np.array([period])
    resp_u, resp_v, resp_a = sdof.response_series(motion=rec, dt=dt, periods=periods, xi=xi)

    k_spring = 4 * np.pi ** 2 * mass / period ** 2
    outputs = get_inelastic_response(mass, k_spring, f_yield, rec, dt, xi=xi, r_post=r_post)
    outputs_elastic = get_inelastic_response(mass, k_spring, f_yield * 100, rec, dt, xi=xi, r_post=r_post)
    ux_opensees = outputs["rel_disp"]
    disp_inelastic_final = ux_opensees[-1]

    time = acc_signal.time
    acc_opensees_elastic = np.interp(time, outputs_elastic["time"], outputs_elastic["rel_accel"]) - rec
    ux_opensees_elastic = np.interp(time, outputs_elastic["time"], outputs_elastic["rel_disp"])
    diff_disp = abs(np.sum(ux_opensees_elastic - resp_u[0]))
    diff_acc = abs(np.sum(acc_opensees_elastic - resp_a[0]))
    assert diff_disp < 1.0e-4
    assert diff_acc < 5.0e-4
    assert np.isclose(disp_inelastic_final, 0.018651987846)


if __name__ == '__main__':
    test_sdof()



from o3seespy.base_model import OpenSeesObject


class SectionBase(OpenSeesObject):
    op_base_type = "section"

    def set_parameter(self, osi, pstr, value, ele, eles):
        from o3seespy import set_parameter
        if ele is not None:
            set_parameter(osi, value=value, eles=[ele], args=[pstr, 1])
        if eles is not None:
            set_parameter(osi, value=value, eles=eles, args=[pstr, 1])


class Elastic2D(SectionBase):
    """
    The Elastic2D Section Class
    
    
    """
    op_type = 'Elastic'

    def __init__(self, osi, e_mod, area, iz, g_mod: float=None, alpha_y: float=None):
        """
        Initial method for Elastic2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        e_mod: float
            Young's modulus
        area: float
            Cross-sectional area of section
        iz: float
            Second moment of area about the local z-axis
        g_mod: float (default=True), optional
            Shear modulus (optional for 2d analysis, required for 3d analysis)
        alpha_y: float (default=True), optional
            Shear shape factor along the local y-axis 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0, g_mod=0.0, alpha_y=0.0)
        """
        self.osi = osi
        self.e_mod = float(e_mod)
        self.area = float(area)
        self.iz = float(iz)
        if g_mod is None:
            self.g_mod = None
        else:
            self.g_mod = float(g_mod)
        if alpha_y is None:
            self.alpha_y = None
        else:
            self.alpha_y = float(alpha_y)
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, self.e_mod, self.area, self.iz]
        special_pms = ['g_mod', 'alpha_y']
        packets = [False, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)

    def set_e_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'E', value, ele, eles)

    def set_a(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'A', value, ele, eles)

    def set_i(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'I', value, ele, eles)


class Elastic3D(SectionBase):
    """
    The Elastic3D Section Class
    
    This command allows the user to construct an ElasticSection. The inclusion of shear deformations is optional. The
    dofs for 2D elastic section are ``[P, Mz]``,for 3D are ``[P,Mz,My,T]``.
    """
    op_type = 'Elastic'

    def __init__(self, osi, e_mod, area, iz, iy, g_mod, jxx, alpha_y: float=None, alpha_z: float=None):
        """
        Initial method for Elastic3D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        e_mod: float
            Young's modulus
        area: float
            Cross-sectional area of section
        iz: float
            Second moment of area about the local z-axis
        iy: float
            Second moment of area about the local y-axis (required for 3d analysis)
        g_mod: float
            Shear modulus (optional for 2d analysis, required for 3d analysis)
        jxx: float
            Torsional moment of inertia of section (required for 3d analysis)
        alpha_y: float (default=True), optional
            Shear shape factor along the local y-axis 
        alpha_z: float (default=True), optional
            Shear shape factor along the local z-axis 

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.section.Elastic3D(osi, e_mod=1.0, area=1.0, iz=1.0, iy=1.0, g_mod=1.0, jxx=1.0, alpha_y=0.0, alpha_z=0.0)
        """
        self.osi = osi
        self.e_mod = float(e_mod)
        self.area = float(area)
        self.iz = float(iz)
        self.iy = float(iy)
        self.g_mod = float(g_mod)
        self.jxx = float(jxx)
        if alpha_y is None:
            self.alpha_y = None
        else:
            self.alpha_y = float(alpha_y)
        if alpha_z is None:
            self.alpha_z = None
        else:
            self.alpha_z = float(alpha_z)
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, self.e_mod, self.area, self.iz, self.iy, self.g_mod, self.jxx]
        special_pms = ['alpha_y', 'alpha_z']
        packets = [False, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        self.to_process(osi)

    def set_e_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'E', value, ele, eles)

    def set_a(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'A', value, ele, eles)

    def set_i(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'I', value, ele, eles)


class Fiber(SectionBase):
    """
    The Fiber Section Class

    This command allows the user to construct a FiberSection object. Each FiberSection object is composed of Fibers,
    with each fiber containing a UniaxialMaterial, an area and a location (y,z). The dofs for 2D section are ``[P,
    Mz]``,for 3D are ``[P,Mz,My,T]``.
    """
    op_type = 'Fiber'

    def __init__(self, osi, gj: float = None, torsion_mat=None):
        """
        Initial method for Fiber

        Supports pre-building

        Parameters
        ----------
        gj: float
            Linear-elastic torsional stiffness assigned to the section
        torsion_mat: obj
            uniaxial_material object assigned to the section for torsional response (can be nonlinear)
        """
        self.osi = osi
        if gj is None:
            self.gj = None
        else:
            self.gj = float(gj)
        self.torsion_mat = torsion_mat
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag]
        if getattr(self, 'gj') is not None:
            self._parameters += ['-GJ', self.gj]
        if getattr(self, 'torsion_mat') is not None:
            self._parameters += ['-torsion', self.torsion_mat.tag]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class Fiber(SectionBase):
    """
    The Fiber Section Class

    This command allows the user to construct a FiberSection object. Each FiberSection object is composed of Fibers,
    with each fiber containing a UniaxialMaterial, an area and a location (y,z). The dofs for 2D section are ``[P,
    Mz]``,for 3D are ``[P,Mz,My,T]``.
    """
    op_type = 'Fiber'

    def __init__(self, osi, gj: float = None, torsion_mat=None):
        """
        Initial method for Fiber

        Supports pre-building

        Parameters
        ----------
        gj: float
            Linear-elastic torsional stiffness assigned to the section
        torsion_mat: obj
            uniaxial_material object assigned to the section for torsional response (can be nonlinear)
        """
        self.osi = osi
        if gj is None:
            self.gj = None
        else:
            self.gj = float(gj)
        self.torsion_mat = torsion_mat
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag]
        if getattr(self, 'gj') is not None:
            self._parameters += ['-GJ', self.gj]
        if getattr(self, 'torsion_mat') is not None:
            self._parameters += ['-torsion', self.torsion_mat.tag]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)


class Fiber(SectionBase):
    """
    The Fiber Section Class

    This command allows the user to construct a FiberSection object. Each FiberSection object is composed of Fibers,
    with each fiber containing a UniaxialMaterial, an area and a location (y,z). The dofs for 2D section are ``[P,
    Mz]``,for 3D are ``[P,Mz,My,T]``.
    """
    op_type = 'Fiber'

    def __init__(self, osi, gj: float = None, torsion_mat=None):
        """
        Initial method for Fiber

        Supports pre-building

        Parameters
        ----------
        gj: float
            Linear-elastic torsional stiffness assigned to the section
        torsion_mat: obj
            uniaxial_material object assigned to the section for torsional response (can be nonlinear)
        """
        self.osi = osi
        if gj is None:
            self.gj = None
        else:
            self.gj = float(gj)
        self.torsion_mat = torsion_mat
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag]
        if getattr(self, 'gj') is not None:
            self._parameters += ['-GJ', self.gj]
        if getattr(self, 'torsion_mat') is not None:
            self._parameters += ['-torsion', self.torsion_mat.tag]
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)



class FiberThermal(SectionBase):
    """
    The FiberThermal Section Class
    
    This command create a FiberSectionThermal object.The dofs for 2D section are ``[P, Mz]``,for 3D are ``[P,Mz,My]``...
    note::#. The commands below should be called after the section command to generate all the fibers in the section.#. The
    patch and layer commands can be used to generate multiple fibers in a single command.Commands to generate all
    fibers:#. :doc:`fiber`#. :doc:`patch`#. :doc:`layer`
    """
    op_type = 'FiberThermal'

    def __init__(self, osi, gj=None):
        """
        Initial method for FiberThermal

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        gj: None, optional
            

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.section.FiberThermal(osi, gj=0.0)
        """
        self.osi = osi
        self.gj = gj
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag]
        if getattr(self, 'gj') is not None:
            self._parameters += ['-GJ', self.gj]
        self.to_process(osi)


class NDFiber(SectionBase):
    """
    The NDFiber Section Class
    
    This commnand allows the user to construct an NDFiberSection object. Each NDFiberSection object is composed of
    NDFibers, with each fiber containing an NDMaterial, an area and a location (y,z). The NDFiberSection works for 2D
    and 3D frame elements and it queries the NDMaterial of each fiber for its axial and shear stresses. In 2D,
    stress components 11 and 12 are obtained from each fiber in order to provide stress resultants for axial
    force, bending moment, and shear ``[P, Mz, Vy]``. Stress components 11, 12, and 13 lead to all six
    stress resultants in 3D ``[P, Mz, Vy, My, Vz, T]``.The NDFiberSection works with any NDMaterial
    via wrapper classes that perform static condensation of the stress vector down to the 11, 12,
    and 13 components, or via concrete NDMaterial subclasses that implement the appropriate fiber stress conditions.
    """
    op_type = 'NDFiber'

    def __init__(self, osi):
        """
        Initial method for NDFiber

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.section.NDFiber(osi)
        """
        self.osi = osi
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag]
        self.to_process(osi)


class WFSection2D(SectionBase):
    """
    The WFSection2D Section Class
    
    This command allows the user to construct a WFSection2d object, which is an encapsulated fiber representation of a
    wide flange steel section appropriate for plane frame analysis.
    """
    op_type = 'WFSection2d'

    def __init__(self, osi, mat, d, tw, bf, tf, nfw, nff):
        """
        Initial method for WFSection2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mat: obj
            Object of uniaxial_material assigned to each fiber
        d: float
            Section depth
        tw: float
            Web thickness
        bf: float
            Flange width
        tf: float
            Flange thickness
        nfw: float
            Number of fibers in the web
        nff: float
            Number of fibers in each flange

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> mat = o3.uniaxial_material.Elastic(osi, 1.0)
        >>> o3.section.WFSection2D(osi, mat=mat, d=1.0, tw=1.0, bf=1.0, tf=1.0, nfw=1.0, nff=1.0)
        """
        self.osi = osi
        self.mat = mat
        self.d = float(d)
        self.tw = float(tw)
        self.bf = float(bf)
        self.tf = float(tf)
        self.nfw = float(nfw)
        self.nff = float(nff)
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, self.mat.tag, self.d, self.tw, self.bf, self.tf, self.nfw, self.nff]
        self.to_process(osi)


class RCSection2D(SectionBase):
    """
    The RCSection2D Section Class
    
    This command allows the user to construct an RCSection2d object, which is an encapsulated fiber representation of a
    rectangular reinforced concrete section with core and confined regions of concrete and single top and bottom layers of
    reinforcement appropriate for plane frame analysis.
    """
    op_type = 'RCSection2d'

    def __init__(self, osi, core_mat, cover_mat, steel_mat, d, b, cover_depth, atop, abot, aside, nfcore, nfcover, nfs):
        """
        Initial method for RCSection2D

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        core_mat: obj
            Object of uniaxial_material assigned to each fiber in the core region
        cover_mat: obj
            Object of uniaxial_material assigned to each fiber in the cover region
        steel_mat: obj
            Object of uniaxial_material assigned to each reinforcing bar
        d: float
            Section depth
        b: float
            Section width
        cover_depth: float
            Cover depth (assumed uniform around perimeter)
        atop: float
            Area of reinforcing bars in top layer
        abot: float
            Area of reinforcing bars in bottom layer
        aside: float
            Area of reinforcing bars on intermediate layers
        nfcore: float
            Number of fibers through the core depth
        nfcover: float
            Number of fibers through the cover depth
        nfs: float
            Number of bars on the top and bottom rows of reinforcement (nfs-2 bars will be placed on the side rows)

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> core_mat = o3.uniaxial_material.Concrete01(osi, -6.0, -0.004, -5.0, -0.014)
        >>> cover_mat = o3.uniaxial_material.Concrete01(osi, -5.0, -0.002, 0.0, -0.006)
        >>> steel_mat = o3.uniaxial_material.Steel01(osi, 60.0, 30000.0, 0.01)
        >>> o3.section.RCSection2D(osi, core_mat=core_mat, cover_mat=cover_mat, steel_mat=steel_mat, d=1.0, b=1.0, cover_depth=1.0, atop=1.0, abot=1.0, aside=1.0, nfcore=1.0, nfcover=1.0, nfs=1.0)
        """
        self.osi = osi
        self.core_mat = core_mat
        self.cover_mat = cover_mat
        self.steel_mat = steel_mat
        self.d = float(d)
        self.b = float(b)
        self.cover_depth = float(cover_depth)
        self.atop = float(atop)
        self.abot = float(abot)
        self.aside = float(aside)
        self.nfcore = float(nfcore)
        self.nfcover = float(nfcover)
        self.nfs = float(nfs)
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, self.core_mat.tag, self.cover_mat.tag, self.steel_mat.tag, self.d, self.b, self.cover_depth, self.atop, self.abot, self.aside, self.nfcore, self.nfcover, self.nfs]
        self.to_process(osi)


class RCCircularSection(SectionBase):
    """
    The RCCircularSection Section Class
    
    This command allows the user to construct an RCCircularSection object, which is an encapsulated fiber representation
    of a circular reinforced concrete section with core and confined regions of concrete.
    """
    op_type = 'RCCircularSection'

    def __init__(self, osi, core_mat, cover_mat, steel_mat, d, cover_depth, a_s, nrings_core, nrings_cover, newedges, nsteel, gj: float=None):
        """
        Initial method for RCCircularSection

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        core_mat: obj
            Object of uniaxial_material assigned to each fiber in the core region
        cover_mat: obj
            Object of uniaxial_material assigned to each fiber in the cover region
        steel_mat: obj
            Object of uniaxial_material assigned to each reinforcing bar
        d: float
            Section radius
        cover_depth: float
            Cover depth (assumed uniform around perimeter)
        a_s: float
            Area of reinforcing bars 
        nrings_core: int
            Number of fibers through the core depth
        nrings_cover: int
            Number of fibers through the cover depth
        newedges: int
            Number of fibers through the edges
        nsteel: int
            Number of fibers through the steels
        gj: float, optional
            Gj stiffness

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> core_mat = o3.uniaxial_material.Concrete01(osi, -6.0, -0.004, -5.0, -0.014)
        >>> cover_mat = o3.uniaxial_material.Concrete01(osi, -5.0, -0.002, 0.0, -0.006)
        >>> steel_mat = o3.uniaxial_material.Steel01(osi, 60.0, 30000.0, 0.01)
        >>> o3.section.RCCircularSection(osi, core_mat=core_mat, cover_mat=cover_mat, steel_mat=steel_mat, d=1.0,
        >>>                              cover_depth=0.10, a_s=0.1, nrings_core=2, nrings_cover=2, newedges=2, nsteel=4, gj=0.0)
        """
        self.osi = osi
        self.core_mat = core_mat
        self.cover_mat = cover_mat
        self.steel_mat = steel_mat
        self.d = float(d)
        self.cover_depth = float(cover_depth)
        self.a_s = float(a_s)
        self.nrings_core = int(nrings_core)
        self.nrings_cover = int(nrings_cover)
        self.newedges = int(newedges)
        self.nsteel = int(nsteel)
        if gj is None:
            self.gj = None
        else:
            self.gj = float(gj)
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, self.core_mat.tag, self.cover_mat.tag, self.steel_mat.tag, self.d, self.cover_depth, self.a_s, self.nrings_core, self.nrings_cover, self.newedges, self.nsteel]
        if getattr(self, 'gj') is not None:
            self._parameters += ['-GJ', self.gj]
        self.to_process(osi)


class Parallel(SectionBase):
    """
    The Parallel Section Class
    
    Connect sections in parallel.
    """
    op_type = 'Parallel'

    def __init__(self, osi, secs):
        """
        Initial method for Parallel

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        secs: list
            Objects of of predefined sections.

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> secs = [o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0, g_mod=0.0, alpha_y=0.0),
        >>>          o3.section.Elastic2D(osi, e_mod=1.0, area=1.0, iz=1.0, g_mod=0.0, alpha_y=0.0)]
        >>> o3.section.Parallel(osi, secs)
        """
        self.osi = osi
        self.secs = [x.tag for x in secs]
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, *self.secs]
        self.to_process(osi)


class Aggregator(SectionBase):
    """
    The Aggregator Section Class

    This command is used to construct a SectionAggregator object which aggregates groups previously-defined
    UniaxialMaterial objects into a single section force-deformation model. Each UniaxialMaterial object
    represents the section force-deformation response for a particular section degree-of-freedom (dof).
    There is no interaction between responses in different dof directions. The aggregation can include
    one previously defined section.
    """
    op_type = 'Aggregator'

    def __init__(self, osi, mats, section=None):
        """
        Initial method for Aggregator

        Parameters
        ----------
        mats: list
            List of mat objs and dofs of previously-defined uniaxialmaterial objects, ``mats =
            [[mattag1,dof1],[mattag2,dof2],...]`` the force-deformation quantity to be modeled by this
            section object. one of the following section dof may be used: * ``'p'`` axial
            force-deformation * ``'mz'`` moment-curvature about section local z-axis *
            ``'vy'`` shear force-deformation along section local y-axis * ``'my'``
            moment-curvature about section local y-axis * ``'vz'`` shear
            force-deformation along section local z-axis * ``'t'`` torsion force-deformation
        section: obj
            Tag of previously-defined section object to which the uniaxialmaterial objects are aggregated as additional
            force-deformation relationships (optional)
        """
        self.mats = []
        for i, mat in enumerate(mats):
            self.mats.append(mats[i][0].tag)
            self.mats.append(mats[i][1])
        self.section = section
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, *self.mats]
        if getattr(self, 'section') is not None:
            self._parameters += ['-section', self.section.tag]
        self.to_process(osi)



class Uniaxial(SectionBase):
    """
    The Uniaxial Section Class
    
    This command is used to construct a UniaxialSection object which uses a previously-defined UniaxialMaterial object
    to represent a single section force-deformation response quantity.
    """
    op_type = 'Uniaxial'

    def __init__(self, osi, mat, quantity):
        """
        Initial method for Uniaxial

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mat: obj
            Object of uniaxial material
        quantity: str
            The force-deformation quantity to be modeled by this section object. one of the following section dof may be
            used: * ``'p'`` axial force-deformation * ``'mz'`` moment-curvature about section local z-axis * ``'vy'`` shear
            force-deformation along section local y-axis * ``'my'`` moment-curvature about section local y-axis * ``'vz'``
            shear force-deformation along section local z-axis * ``'t'`` torsion force-deformation

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> mat = o3.uniaxial_material.Elastic(osi, 1.0)
        >>> o3.section.Uniaxial(osi, mat=mat, quantity='P')
        """
        self.osi = osi
        self.mat = mat
        self.quantity = quantity
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, self.mat.tag, self.quantity]
        self.to_process(osi)


class ElasticMembranePlateSection(SectionBase):
    """
    The ElasticMembranePlateSection Section Class
    
    This command allows the user to construct an ElasticMembranePlateSection object, which is an isotropic section
    appropriate for plate and shell analysis.
    """
    op_type = 'ElasticMembranePlateSection'

    def __init__(self, osi, e_mod, nu, h, rho):
        """
        Initial method for ElasticMembranePlateSection

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        e_mod: float
            Young's modulus
        nu: float
            Poisson's ratio
        h: float
            Depth of section
        rho: float
            Mass density

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.section.ElasticMembranePlateSection(osi, e_mod=1.0, nu=1.0, h=1.0, rho=1.0)
        """
        self.osi = osi
        self.e_mod = float(e_mod)
        self.nu = float(nu)
        self.h = float(h)
        self.rho = float(rho)
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, self.e_mod, self.nu, self.h, self.rho]
        self.to_process(osi)


class PlateFiber(SectionBase):
    """
    The PlateFiber Section Class
    
    This command allows the user to construct a MembranePlateFiberSection object, which is a section that numerically
    integrates through the plate thickness with "fibers" and is appropriate for plate and shell analysis.
    """
    op_type = 'PlateFiber'

    def __init__(self, osi, mat, h):
        """
        Initial method for PlateFiber

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        mat: obj
            Ndmaterial object to be assigned to each fiber
        h: float
            Plate thickness

        Examples
        --------
        >>> import o3seespy as o3
        >>> # Example is currently not working
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> mat = o3.uniaxial_material.Elastic(osi, 1.0)
        >>> o3.section.PlateFiber(osi, mat=mat, h=1.0)
        """
        self.osi = osi
        self.mat = mat
        self.h = float(h)
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, self.mat.tag, self.h]
        self.to_process(osi)


class Bidirectional(SectionBase):
    """
    The Bidirectional Section Class
    
    This command allows the user to construct a Bidirectional section, which is a stress-resultant plasticity model of
    two coupled forces. The yield surface is circular and there is combined isotropic and kinematic hardening.
    """
    op_type = 'Bidirectional'

    def __init__(self, osi, e_mod, fy, hiso, hkin, code1='Vy', code2='P'):
        """
        Initial method for Bidirectional

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        e_mod: float
            Elastic modulus
        fy: float
            Yield force
        hiso: float
            Isotropic hardening modulus
        hkin: float
            Kinematic hardening modulus
        code1: str, optional
            Section force code for direction 1 
        code2: str, optional
            Section force code for direction 2  one of the following section code may be used: * ``'p'`` axial
            force-deformation * ``'mz'`` moment-curvature about section local z-axis * ``'vy'`` shear force-deformation
            along section local y-axis * ``'my'`` moment-curvature about section local y-axis * ``'vz'`` shear
            force-deformation along section local z-axis * ``'t'`` torsion force-deformation

        Examples
        --------
        >>> import o3seespy as o3
        >>> osi = o3.OpenSeesInstance(ndm=2)
        >>> o3.section.Bidirectional(osi, e_mod=1.0, fy=1.0, hiso=1.0, hkin=1.0, code1='Vy', code2='P')
        """
        self.osi = osi
        self.e_mod = float(e_mod)
        self.fy = float(fy)
        self.hiso = float(hiso)
        self.hkin = float(hkin)
        self.code1 = code1
        self.code2 = code2
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, self.e_mod, self.fy, self.hiso, self.hkin, self.code1, self.code2]
        self.to_process(osi)


class Isolator2spring(SectionBase):
    """
    The Isolator2spring Section Class
    
    This command is used to construct an Isolator2spring section object, which represents the buckling behavior of an
    elastomeric bearing for two-dimensional analysis in the lateral and vertical plane. An Isolator2spring section
    represents the resultant force-deformation behavior of the bearing, and should be used with a
    zeroLengthSection element. The bearing should be constrained against rotation.
    """
    op_type = 'Isolator2spring'

    def __init__(self, osi, tol, k1, fyo, k2o, kvo, hb, pe, po=0.0):
        """
        Initial method for Isolator2spring

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        tol: float
            Tolerance for convergence of the element state. suggested value: e-12 to e-10. opensees will warn if
            convergence is not achieved, however this usually does not prevent global convergence.
        k1: float
            Initial stiffness for lateral force-deformation
        fyo: float
            Nominal yield strength for lateral force-deformation
        k2o: float
            Nominal postyield stiffness for lateral force-deformation
        kvo: float
            Nominal stiffness in the vertical direction
        hb: float
            Total height of elastomeric bearing
        pe: float
            Euler buckling load for the bearing
        po: float, optional
            Axial load at which nominal yield strength is achieved 
        """
        self.osi = osi
        self.tol = float(tol)
        self.k1 = float(k1)
        self.fyo = float(fyo)
        self.k2o = float(k2o)
        self.kvo = float(kvo)
        self.hb = float(hb)
        self.pe = float(pe)
        self.po = float(po)
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, self.tol, self.k1, self.fyo, self.k2o, self.kvo, self.hb, self.pe, self.po]
        self.to_process(osi)


class LayeredShell(SectionBase):
    """
    The LayeredShell Section Class

    This command will create the section of the multi-layer shell element, including the multi-dimensional concrete,
    reinforcement material and the corresponding thickness.
    """
    op_type = 'LayeredShell'

    def __init__(self, osi, mats):
        """
        Initial method for LayeredShell

        Parameters
        ----------
        mats: list
            A list of material objs and thickness, ``[[mat1,thk1], ..., [mat2,thk2]]``
        """
        self.mats = []
        for i, mat in enumerate(mats):
            # self.mats.append([mats[i][0].tag, mats[i][1]])
            self.mats.append(mats[i][0].tag)
            self.mats.append(mats[i][1])

        self.n_layers = int(len(self.mats))
        osi.n_sect += 1
        self._tag = osi.n_sect
        self._parameters = [self.op_type, self._tag, self.n_layers, *self.mats]
        self.to_process(osi)



import o3seespy as o3  # for testing only


def test_truss():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat = o3.uniaxial_material.Elastic(osi, 1.0)
    o3.element.Truss(osi, ele_nodes=ele_nodes, big_a=1.0, mat=mat, rho=1.0, c_flag=1.0, r_flag=1.0)


def test_corot_truss():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [1, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat = o3.uniaxial_material.Elastic(osi, 1.0)
    o3.element.CorotTruss(osi, ele_nodes=ele_nodes, big_a=1.0, mat=mat, rho=1.0, c_flag=1.0, r_flag=1.0)




import o3seespy as o3  # for testing only
import pytest

def test_std_brick():
    osi = o3.OpenSeesInstance(ndm=3)
    coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    o3.element.StdBrick(osi, ele_nodes=ele_nodes, mat=mat, b1=1.0, b2=1.0, b3=1.0)


def test_bbar_brick():
    osi = o3.OpenSeesInstance(ndm=3)
    coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    o3.element.BbarBrick(osi, ele_nodes=ele_nodes, mat=mat, b1=1.0, b2=1.0, b3=1.0)


@pytest.mark.skip()
def test_n20node_brick():
    osi = o3.OpenSeesInstance(ndm=3, ndf=6)
    coords = [[0, 0, 0], [-1, 0, 0], [-1, -1, 0], [0, -1, 0],
              [0, 0, -1], [-1, 0, -1], [-1, -1, -1], [0, -1, -1],
              [-0.5, 0, 0], [-1, -0.5, 0], [-0.5, -1, 0], [0, -0.5, 0],
              [-0.5, 0, -1], [-1, -0.5, -1], [-0.5, -1, -1], [0, -0.5, -1],
              [0, 0, -0.5], [-1, 0, -0.5], [-1, -1, -0.5], [0, -1, -0.5],
              ]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    o3.element.N20NodeBrick(osi, ele_nodes=ele_nodes, mat=mat, bf1=1.0, bf2=1.0, bf3=1.0, mass_den=1.0)


def test_ss_pbrick():
    osi = o3.OpenSeesInstance(ndm=3)
    coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    o3.element.SSPbrick(osi, ele_nodes=ele_nodes, mat=mat, b1=1.0, b2=1.0, b3=1.0)




import o3seespy as o3  # for testing only
import pytest


@pytest.mark.skip()
def test_line():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0], [0, 0]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(2)]
    o3.mesh.Line(osi, numnodes=1, ndtags=ele_nodes, id=1, ndf=1, meshsize=1.0, ele_type='', ele_args=None)


@pytest.mark.skip()
def test_tri():
    osi = o3.OpenSeesInstance(ndm=2)
    ltags = [1, 1]
    o3.mesh.Tri(osi, numlines=1, ltags=ltags, id=1, ndf=1, meshsize=1.0, ele_type='', ele_args=None)


@pytest.mark.skip()
def test_quad():
    osi = o3.OpenSeesInstance(ndm=2)
    ltags = [1, 1]
    o3.mesh.Quad(osi, numlines=1, ltags=ltags, id=1, ndf=1, meshsize=1.0, ele_type='', ele_args=None)


@pytest.mark.skip()
def test_tet():
    osi = o3.OpenSeesInstance(ndm=2)
    mtags = [1, 1]
    o3.mesh.Tet(osi, nummesh=1, mtags=mtags, id=1, ndf=1, meshsize=1.0, ele_type='', ele_args=None)


@pytest.mark.skip()
def test_partvel():
    osi = o3.OpenSeesInstance(ndm=2)
    p_args = [1.0, 1.0]
    vel0 = [1.0, 1.0]
    o3.mesh.Partvel(osi, otype="string", p_args=p_args, ele_type='', ele_args=None, vel0=vel0)


@pytest.mark.skip()
def test_partpressure():
    osi = o3.OpenSeesInstance(ndm=2)
    p_args = [1.0, 1.0]
    o3.mesh.Partpressure(osi, otype="string", p_args=p_args, ele_type='', ele_args=None, p0=1.0)


@pytest.mark.skip()
def test_bgwave():
    osi = o3.OpenSeesInstance(ndm=2)
    lower = [1.0, 1.0]
    upper = [1.0, 1.0]
    locations = [1.0, 1.0]
    o3.mesh.Bgwave(osi, lower=lower, upper=upper, tol=1.0, meshtol=1.0, wavefilename="string", numl=1, locations=locations, numsub=1)


@pytest.mark.skip()
def test_bgstructure():
    osi = o3.OpenSeesInstance(ndm=2)
    lower = [1.0, 1.0]
    upper = [1.0, 1.0]
    o3.mesh.Bgstructure(osi, lower=lower, upper=upper, tol=1.0, meshtol=1.0, id=1, numnodes=1, snodes=1)


@pytest.mark.skip()
def test_bglarge_size():
    osi = o3.OpenSeesInstance(ndm=2)
    lower = [1.0, 1.0]
    upper = [1.0, 1.0]
    llower = [1.0, 1.0]
    lupper = [1.0, 1.0]
    o3.mesh.BglargeSize(osi, lower=lower, upper=upper, tol=1.0, meshtol=1.0, level=1, llower=llower, lupper=lupper)




from o3seespy.base_model import OpenSeesObject


class AlgorithmBase(OpenSeesObject):
    op_base_type = "algorithm"


class Linear(AlgorithmBase):
    op_type = "Linear"

    def __init__(self, osi, secant=False, initial=False, factor_once=False):
        self.osi = osi
        self.secant = secant
        self.initial = initial
        self.factor_once = factor_once
        self._parameters = [self.op_type, self.secant, self.initial, self.factor_once]
        self.to_process(osi)


class Newton(AlgorithmBase):
    op_type = "Newton"

    def __init__(self, osi, secant=False, initial=False, initial_then_current=False):
        self.osi = osi
        self.secant = secant
        self.initial = initial
        self.initial_then_current = initial_then_current
        self._parameters = [self.op_type, self.secant, self.initial, self.initial_then_current]
        self.to_process(osi)


class SecantNewton(AlgorithmBase):
    op_type = "SecantNewton"

    def __init__(self, osi, iterate='current', increment='current', maxDim=3):
        self.osi = osi
        self.iterate = iterate
        self.increment = increment
        self.maxDim = maxDim
        self._parameters = [self.op_type, self.iterate, self.increment, self.maxDim]
        self.to_process(osi)


class RaphsonNewton(AlgorithmBase):
    op_type = "RaphsonNewton"

    def __init__(self, osi, iterate='current', increment='current'):
        self.osi = osi
        self.iterate = iterate
        self.increment = increment
        self._parameters = [self.op_type, self.iterate, self.increment]
        self.to_process(osi)


class KrylovNewton(AlgorithmBase):
    op_type = "KrylovNewton"

    def __init__(self, osi, tang_inter='current', tang_incr='current', max_inter=3):
        """

        Parameters
        ----------
        osi
        tang_inter: str
            options are: 'current', 'initial', 'noTangent'
        tang_incr
        max_inter
        """
        self.osi = osi
        self.tang_inter = tang_inter
        self.tang_incr = tang_incr
        self.max_inter = max_inter
        self._parameters = [self.op_type, self.tang_inter, self.tang_incr, self.max_inter]
        self.to_process(osi)


class ModifiedNewton(AlgorithmBase):
    op_type = "ModifiedNewton"

    def __init__(self, osi, secant=False, initial=False):
        """

        Parameters
        ----------
        osi
        secant: bool
            Flag to indicate to use secant stiffness.
        initial: bool
            Flag to indicate to use initial stiffness
        max_inter
        """
        self.osi = osi
        self.secant = secant
        self.initial = initial
        self._parameters = [self.op_type, self.secant, self.initial]
        self.to_process(osi)


import o3seespy as o3  # for testing only
import pytest

def test_quad_up():
    osi = o3.OpenSeesInstance(ndm=2)
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    o3.element.QuadUP(osi, ele_nodes=ele_nodes, thick=1.0, mat=mat, bulk=1.0, fmass=1.0, h_perm=1.0, v_perm=1.0, b1=0, b2=0, t=0)


@pytest.mark.skip()
def test_brick_up():
    osi = o3.OpenSeesInstance(ndm=3, ndf=6)
    coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    o3.element.BrickUP(osi, ele_nodes=ele_nodes, mat=mat, bulk=1.0, fmass=1.0, perm_x=1.0, perm_y=1.0, perm_z=1.0, b_x=0, b_y=0, b_z=0)


def test_bbar_quad_up():
    osi = o3.OpenSeesInstance(ndm=2)
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    coords = [[0, 0], [1, 0], [1, 1], [0, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    o3.element.BbarQuadUP(osi, ele_nodes=ele_nodes, thick=1.0, mat=mat, bulk=1.0, fmass=1.0, h_perm=1.0, v_perm=1.0, b1=0, b2=0, t=0)


@pytest.mark.skip()
def test_bbar_brick_up():
    osi = o3.OpenSeesInstance(ndm=3)
    coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    o3.element.BbarBrickUP(osi, ele_nodes=ele_nodes, mat=mat, bulk=1.0, fmass=1.0, perm_x=1.0, perm_y=1.0, perm_z=1.0, b_x=0, b_y=0, b_z=0)


@pytest.mark.skip()
def test_n94quad_up():
    osi = o3.OpenSeesInstance(ndm=2)
    coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1], [0.5, 0.5, 0.5]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    o3.element.N94QuadUP(osi, ele_nodes=ele_nodes, thick=1.0, mat=mat, bulk=1.0, fmass=1.0, h_perm=1.0, v_perm=1.0, b1=0, b2=0)


@pytest.mark.skip()
def test_n208brick_up():
    osi = o3.OpenSeesInstance(ndm=3)
    coords = [[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1]]
    ele_nodes = [o3.node.Node(osi, *coords[x]) for x in range(len(coords))]
    mat = o3.nd_material.ElasticIsotropic(osi, 1, 0.45)
    o3.element.N208BrickUP(osi, ele_nodes=ele_nodes, mat=mat, bulk=1.0, fmass=1.0, perm_x=1.0, perm_y=1.0, perm_z=1.0, b_x=0, b_y=0, b_z=0)




from o3seespy.command.nd_material.base_material import NDMaterialBase


class PressureIndependMultiYield(NDMaterialBase):
    op_type = "PressureIndependMultiYield"

    def __init__(self, osi,  nd, rho, g_mod_ref, bulk_mod_ref, cohesion, peak_strain, phi=0.,
                 p_ref=100., d=0., n_surf=20, strains=None, ratios=None):
        r"""
        PressureIndependMultiYield material

        Parameters
        ==========
        matTag: int
            integer tag identifying material
        nd: float
            Number of dimensions, 2 for plane-strain, and 3 for 3D analysis.
       rho: float
            Saturated soil mass density.
       g_mod_ref, float
            (:math:`G_0`) Reference low-strain shear modulus,
            specified at a reference mean effective confining pressure (`p_ref`).
       bulk_mod_ref: float
            (:math:`B_r`) Reference bulk modulus,
            specified at a reference mean effective confining pressure (`p_ref`)).
       cohesion: float
            (:math:`c`) Apparent cohesion at zero effective confinement.
       peak_strain: float
            (:math:`\gamma_{max}`) An octahedral shear strain at
            which the maximum shear strength is reached,
            specified at a reference mean effective confining
            pressure refPress of p'r (see below).
       phi: float
            (:math:`phi`) Friction angle at peak shear
            strength in degrees, optional (default is 0.0).
       p_ref: float
            (:math:`p'_ref`) Reference mean effective confining pressure at which
                          :math:`G_r`, :math:`B_r`, and :math:`\gamma_{max}`
                          are defined, optional (default is 100. kPa).
       d: float
            (:math:`d`) A positive constant defining variations
                        of :math:`G` and :math:`B` as a function of
                          instantaneous effective
                          confinement :math:`p'` (default is 0.0)

                          :math:`G=G_r(\frac{p'}{p'_ref})^d`

                          :math:`B=B_r(\frac{p'}{p'_ref})^d`

                          If :math:`\phi=0`, :math:`d` is reset to 0.0.

       n_surf: float, optional
            Number of yield surfaces, optional (must be less
            than 40, default is 20). The surfaces are generated
            based on the hyperbolic relation.
       strains: list
            Instead of automatic surfaces generation,
            you can define yield surfaces directly based on
            desired shear modulus reduction curve.
            provide a list of shear strains and corresponding shear modulus reduction ratios (`ratios`)
        ratios: list
            Instead of automatic surfaces generation,
            you can define yield surfaces directly based on
            desired shear modulus reduction curve.
            provide a list of shear strains and corresponding shear modulus reduction ratios (`ratios`)
        """
        self.osi = osi
        self.nd = nd
        self.rho = float(rho)
        self.g_mod_ref = float(g_mod_ref)
        self.bulk_mod_ref = float(bulk_mod_ref)
        self.cohesion = float(cohesion)
        self.peak_strain = float(peak_strain)
        self.phi = float(phi)
        self.p_ref = float(p_ref)
        self.d = float(d)
        assert n_surf < 40
        self.n_surf = int(n_surf)
        if strains is not None:
            assert len(strains) == len(ratios)
            yield_surf = []
            for i in range(len(strains)):
                yield_surf.append(ratios[i])
                yield_surf.append(strains[i])
            self.yield_surf = yield_surf
        else:
            self.yield_surf = None
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat

        self._parameters = [self.op_type, self._tag, self.nd, self.rho, self.g_mod_ref, self.bulk_mod_ref,
                            self.cohesion, self.peak_strain, self.phi, self.p_ref, self.d]

        if self.yield_surf is not None:
            self._parameters.append(self.n_surf)  # from docs 'add a minus sign in front of noYieldSurf'
            self._parameters += list(self.yield_surf)

        else:
            # self._keyword_args['noYieldSurf'] = self.no_yield_surf
            self._parameters.append(self.n_surf)
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def update_to_nonlinear(self):
        from o3seespy import update_material_stage
        update_material_stage(self.osi, self, 1)

    def set_nu(self, nu, ele=None, eles=None, adj_g_mod=False):
        if adj_g_mod:
            g_mod = 3 * self.bulk_mod_ref * (1 - 2 * nu) / (2 * (1 + nu))
            self.set_parameter(self.osi, 'shearModulus', g_mod, ele, eles)
        else:
            bulk_mod = 2 * self.g_mod_ref * (1 + nu) / (3 * (1 - 2 * nu))
            self.set_parameter(self.osi, 'bulkModulus', bulk_mod, ele, eles)



class PressureDependMultiYield(NDMaterialBase):
    """
    The PressureDependMultiYield NDMaterial Class

    PressureDependMultiYield material is an elastic-plastic material for simulating the essential response
    characteristics of pressure sensitive soil materials under general loading conditions. Such
    characteristics include dilatancy (shear-induced volume contraction or dilation) and
    non-flow liquefaction (cyclic mobility), typically exhibited in sands or silts during monotonic or cyclic loading.
    """
    op_type = 'PressureDependMultiYield'

    def __init__(self, osi, nd, rho, g_mod_ref, bulk_mod_ref, phi, peak_strain, p_ref,
                 d, pt_ang, con_rate, dil_rates, liquefac, n_surf=20.0, strains=None, ratios=None,
                 e_init=0.6, cs_params=None, c=0.3):
        r"""
        Initial method for PressureDependMultiYield

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        nd: int
            Number of dimensions, 2 for plane-strain, and 3 for 3d analysis.
        rho: float
            Saturated soil mass density.
        g_mod_ref, float
            (:math:`G_0`) Reference low-strain shear modulus,
            specified at a reference mean effective confining pressure (`p_ref`).
       bulk_mod_ref: float
            (:math:`B_r`) Reference bulk modulus,
            specified at a reference mean effective confining pressure (`p_ref`)).
        phi: float
            (:math:`phi`) friction angle at peak shear strength in degrees, optional (default is 0.0).
        peak_strain: float
            (:math:`\\gamma_{max}`) An octahedral shear strain at
            which the maximum shear strength is reached,
            specified at a reference mean effective confining
            pressure refPress of p'r (see below).
        p_ref: float
            (:math:`p'_r`) reference mean effective confining pressure at which :math:`g_r`, :math:`b_r`, and
            :math:`\gamma_{max}` are defined, optional (default is 100. kpa).
        d: float
            (:math:`d`) a positive constant defining variations of :math:`g` and :math:`b` as a function of
            instantaneous effective confinement :math:`p'` (default is 0.0) :math:`g=g_r(\frac{p'}{p'_r})^d`
            :math:`b=b_r(\frac{p'}{p'_r})^d` if :math:`\phi=0`, :math:`d` is reset to 0.0.
        pt_ang: float
            (:math:`\phi_{pt}`) phase transformation angle, in degrees.
        con_rate: float
            A non-negative constant defining the rate of shear-induced volume decrease (contraction) or pore pressure
            buildup. a larger value corresponds to faster contraction rate.
        dil_rates: list
            Non-negative constants defining the rate of shear-induced volume increase (dilation). larger values
            correspond to stronger dilation rate. ``dil_rates = [dilat1, dilat2]``.
        liquefac: list
            Parameters controlling the mechanism of liquefaction-induced perfectly plastic shear strain accumulation,
            i.e., cyclic mobility. set ``liquefac[0] = 0`` to deactivate this mechanism altogether. ``liquefac[0]`` defines the
            effective confining pressure (e.g., 10 kpa in si units or 1.45 psi in english units) below which the mechanism is
            in effect. smaller values should be assigned to denser sands. ``liquefac[1]`` defines the maximum amount of
            perfectly plastic shear strain developed at zero effective confinement during each loading phase. smaller
            values should be assigned to denser sands. ``liquefac[2]`` defines the maximum amount of biased
            perfectly plastic shear strain :math:`\gamma_b` accumulated at each loading phase under biased
            shear loading conditions, as :math:`\gamma_b=liquefac[1]\times liquefac[2]`. typically,
            :math:`liquefac[2]` takes a value between 0.0 and 3.0. smaller values should be
            assigned to denser sands. see the references listed at the end of this chapter for more information.
        n_surf: int, optional
            Number of yield surfaces, optional (must be less than 40, default is 20). the surfaces are generated based
            on the hyperbolic relation defined in note 2 below.
        strains: list
            Instead of automatic surfaces generation,
            you can define yield surfaces directly based on
            desired shear modulus reduction curve.
            provide a list of shear strains and corresponding shear modulus reduction ratios (`ratios`)
        ratios: list
            Instead of automatic surfaces generation,
            you can define yield surfaces directly based on
            desired shear modulus reduction curve.
            provide a list of shear strains and corresponding shear modulus reduction ratios (`ratios`)
        e_init: float, optional
            Initial void ratio, optional (default is 0.6).
        cs_params: list (default=[0.9, 0.02, 0.7, 101.0]), optional
            ``params=[cs1, cs2, cs3, pa]`` defining a straight critical-state line ec in e-p' space. if cs3=0, ec =
            cs1-cs2 log(p'/pa) else (li and wang, jgge, 124(12)), ec = cs1-cs2(p'/pa)cs3 where pa is atmospheric pressure for
            normalization (typically 101 kpa in si units, or 14.65 psi in english units). all four constants are optional
        c: float, optional
            Numerical constant (default value = 0.3 kPa)
        """
        self.osi = osi
        self.nd = int(nd)
        self.rho = float(rho)
        self.g_mod_ref = float(g_mod_ref)
        self.bulk_mod_ref = float(bulk_mod_ref)
        self.peak_strain = float(peak_strain)
        self.phi = float(phi)
        self.p_ref = float(p_ref)
        self.d = float(d)
        self.pt_ang = float(pt_ang)
        self.con_rate = float(con_rate)
        self.dil_rates = dil_rates
        self.liquefac = liquefac
        assert n_surf < 40
        self.n_surf = int(n_surf)
        if strains is not None:
            assert len(strains) == len(ratios)
            yield_surf = []
            for i in range(len(strains)):
                yield_surf.append(ratios[i])
                yield_surf.append(strains[i])
            self.yield_surf = yield_surf
        else:
            self.yield_surf = None
        self.e_init = float(e_init)
        self.cs_params = cs_params
        self.c = float(c)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.nd, self.rho, self.g_mod_ref, self.bulk_mod_ref,
                            self.phi, self.peak_strain, self.p_ref, self.d, self.pt_ang,
                            self.con_rate, *self.dil_rates, *self.liquefac, self.n_surf]
        if self.yield_surf is not None:
            self._parameters += list(self.yield_surf)
        special_pms = ['e_init', 'cs_params', 'c']
        packets = [False, True, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def update_to_nonlinear(self):
        from o3seespy import update_material_stage
        update_material_stage(self.osi, self, 1)

    def set_nu(self, nu, ele=None, eles=None, adj_g_mod=False):
        if adj_g_mod:
            g_mod = 3 * self.bulk_mod_ref * (1 - 2 * nu) / (2 * (1 + nu))
            self.set_parameter(self.osi, 'shearModulus', g_mod, ele, eles)
        else:
            bulk_mod = 2 * self.g_mod_ref * (1 + nu) / (3 * (1 - 2 * nu))
            self.set_parameter(self.osi, 'bulkModulus', bulk_mod, ele, eles)



class PressureDependMultiYield02(NDMaterialBase):
    r"""
    The PressureDependMultiYield02 NDMaterial Class

    PressureDependMultiYield02 material is modified from PressureDependMultiYield material, with:#. additional
    parameters (``contrac[2]`` and ``dilat[2]``) to account for :math:`K_{\sigma}` effect,#. a parameter to
    account for the influence of previous dilation history on subsequent contraction phase
    (``contrac[1]``), and#. modified logic related to permanent shear strain accumulation
    (``liquefac[0]`` and ``liquefac[1]``).
    """
    op_type = 'PressureDependMultiYield02'

    def __init__(self, osi, nd, rho, g_mod_ref, bulk_mod_ref, phi, peak_strain, p_ref,
                 d, pt_ang, con_rates, dil_rates, liquefac=(1., 0.), n_surf=20.0, strains=None, ratios=None,
                 e_init=0.6, cs_params=None, c=0.1):
        r"""
        Initial method for PressureDependMultiYield02

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        nd: int
            Number of dimensions, 2 for plane-strain, and 3 for 3d analysis.
        rho: float
            Saturated soil mass density.
        g_mod_ref, float
            (:math:`G_0`) Reference low-strain shear modulus,
            specified at a reference mean effective confining pressure (`p_ref`).
       bulk_mod_ref: float
            (:math:`B_r`) Reference bulk modulus,
            specified at a reference mean effective confining pressure (`p_ref`)).
        phi: float
            (:math:`phi`) friction angle at peak shear strength in degrees, optional (default is 0.0).
        peak_strain: float
            (:math:`\\gamma_{max}`) An octahedral shear strain at
            which the maximum shear strength is reached,
            specified at a reference mean effective confining
            pressure refPress of p'r (see below).
        p_ref: float
            (:math:`p'_r`) reference mean effective confining pressure at which :math:`g_r`, :math:`b_r`, and
            :math:`\gamma_{max}` are defined, optional (default is 100. kpa).
        d: float
            (:math:`d`) a positive constant defining variations of :math:`g` and :math:`b` as a function of
            instantaneous effective confinement :math:`p'` (default is 0.0) :math:`g=g_r(\frac{p'}{p'_r})^d`
            :math:`b=b_r(\frac{p'}{p'_r})^d` if :math:`\phi=0`, :math:`d` is reset to 0.0.
        pt_ang: float
            (:math:`\phi_{pt}`) phase transformation angle, in degrees.
        con_rates: list
            A list of constants defining contraction behaviour `[contrac1, contrac2, contrac3]`.
             `contrac1`: non-negative constant defining the rate of shear-induced volume decrease (contraction)
             or pore pressure buildup. a larger value corresponds to faster contraction rate.
             `contrac2`: A non-negative constant reflecting dilation history on contraction tendency.
             `contrac3`: A non-negative constant reflecting Kσ effect.
        dil_rates: list
            A list of constants defining contraction behaviour `[dilat1, dilat2, dilat3]`.
            `dilat1` and `dilat2`: Non-negative constants defining the rate of shear-induced volume
            increase (dilation). larger values correspond to stronger dilation rate.
            `dilat3`: A non-negative constant reflecting Kσ effect.
        liquefac: list
            Parameters controlling the mechanism of liquefaction-induced perfectly plastic shear strain accumulation,
            i.e., cyclic mobility. [`liquefac1`, `liquefac2`]. NOTE: Different from `PressureDependMultiYield`.
            `liquefac1`: Damage parameter to define accumulated permanent shear strain as a function of
            dilation history.
            `liquefac2`: Damage parameter to define biased accumulation of permanent shear strain as a
            function of load reversal history.
        n_surf: int, optional
            Number of yield surfaces, optional (must be less than 40, default is 20). the surfaces are generated based
            on the hyperbolic relation defined in note 2 below.
        strains: list
            Instead of automatic surfaces generation,
            you can define yield surfaces directly based on
            desired shear modulus reduction curve.
            provide a list of shear strains and corresponding shear modulus reduction ratios (`ratios`)
        ratios: list
            Instead of automatic surfaces generation,
            you can define yield surfaces directly based on
            desired shear modulus reduction curve.
            provide a list of shear strains and corresponding shear modulus reduction ratios (`ratios`)
        e_init: float, optional
            Initial void ratio, optional (default is 0.6).
        cs_params: list (default=[0.9, 0.02, 0.7, 101.0]), optional
            ``params=[cs1, cs2, cs3, pa]`` defining a straight critical-state line ec in e-p' space. if cs3=0, ec =
            cs1-cs2 log(p'/pa) else (li and wang, jgge, 124(12)), ec = cs1-cs2(p'/pa)cs3 where pa is atmospheric pressure for
            normalization (typically 101 kpa in si units, or 14.65 psi in english units). all four constants are optional
        c: float, optional
            Numerical constant (default value = 0.1 kPa)
        """
        self.osi = osi
        self.nd = int(nd)
        self.rho = float(rho)
        self.g_mod_ref = float(g_mod_ref)
        self.bulk_mod_ref = float(bulk_mod_ref)
        self.peak_strain = float(peak_strain)
        self.phi = float(phi)
        self.p_ref = float(p_ref)
        self.d = float(d)
        self.pt_ang = float(pt_ang)
        self.con_rates = con_rates
        contrac1 = con_rates[0]
        contrac2 = con_rates[1]
        contrac3 = con_rates[2]
        self.dil_rates = dil_rates
        dilat1 = dil_rates[0]
        dilat2 = dil_rates[1]
        dilat3 = dil_rates[2]
        self.liquefac = liquefac
        assert n_surf < 40
        self.n_surf = int(n_surf)
        if strains is not None:
            assert len(strains) == len(ratios)
            yield_surf = []
            for i in range(len(strains)):
                yield_surf.append(ratios[i])
                yield_surf.append(strains[i])
            self.yield_surf = yield_surf
        else:
            self.yield_surf = None
        self.e_init = float(e_init)
        self.cs_params = cs_params
        self.c = float(c)
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.nd, self.rho, self.g_mod_ref, self.bulk_mod_ref,
                            self.phi, self.peak_strain, self.p_ref, self.d, self.pt_ang,
                            contrac1, contrac3, dilat1, dilat3, self.n_surf]
        if self.yield_surf is not None:
            self._parameters += list(self.yield_surf)

        self._parameters += [contrac2, dilat2]
        special_pms = ['liquefac', 'e_init', 'cs_params', 'c']
        packets = [True, False, True, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def update_to_nonlinear(self):
        from o3seespy import update_material_stage
        update_material_stage(self.osi, self, 1)

    def set_nu(self, nu, ele=None, eles=None, adj_g_mod=False):
        if adj_g_mod:
            g_mod = 3 * self.bulk_mod_ref * (1 - 2 * nu) / (2 * (1 + nu))
            self.set_parameter(self.osi, 'shearModulus', g_mod, ele, eles)
        else:
            bulk_mod = 2 * self.g_mod_ref * (1 + nu) / (3 * (1 - 2 * nu))
            self.set_parameter(self.osi, 'bulkModulus', bulk_mod, ele, eles)



class PressureDependMultiYield03(NDMaterialBase):
    """
    The PressureDependMultiYield03 NDMaterial Class
    
    The reference for PressureDependMultiYield03 material: Khosravifar, A., Elgamal, A., Lu, J., and Li, J. [2018]. "A
    3D model for earthquake-induced liquefaction triggering and post-liquefaction response." Soil Dynamics and Earthquake
    Engineering, 110, 43-52)PressureDependMultiYield03 is modified from PressureDependMultiYield02 material to comply
    with the established guidelines on the dependence of liquefaction triggering to the number of loading cycles,
    effective overburden stress (Kσ), and static shear stress (Kα).The explanations of parametersSee `notes
    <http://opensees.berkeley.edu/wiki/index.php/PressureDependMultiYield02_Material>`_
    """
    op_type = 'PressureDependMultiYield03'

    def __init__(self, osi, nd, rho, ref_shear_modul, ref_bulk_modul, friction_ang, peak_shear_stra, ref_press, press_depend_coe, pt_ang, ca, cb, cc, cd, ce, da, db, dc, no_yield_surf=20.0, yield_surf: float=None, liquefac1=1, liquefac2=0., pa=101, s0=1.73):
        """
        Initial method for PressureDependMultiYield03

        Parameters
        ----------
        osi: o3seespy.OpenSeesInstance
        nd: None
            
        rho: None
            
        ref_shear_modul: None
            
        ref_bulk_modul: None
            
        friction_ang: None
            
        peak_shear_stra: None
            
        ref_press: None
            
        press_depend_coe: None
            
        pt_ang: None
            
        ca: None
            
        cb: None
            
        cc: None
            
        cd: None
            
        ce: None
            
        da: None
            
        db: None
            
        dc: None
            
        no_yield_surf: None, optional
            
        yield_surf: None (default=True), optional
            
        liquefac1: None, optional
            
        liquefac2: None, optional
            
        pa: None, optional
            
        s0: None, optional
            
        """
        self.osi = osi
        self.nd = nd
        self.rho = rho
        self.ref_shear_modul = ref_shear_modul
        self.ref_bulk_modul = ref_bulk_modul
        self.friction_ang = friction_ang
        self.peak_shear_stra = peak_shear_stra
        self.ref_press = ref_press
        self.press_depend_coe = press_depend_coe
        self.pt_ang = pt_ang
        self.ca = ca
        self.cb = cb
        self.cc = cc
        self.cd = cd
        self.ce = ce
        self.da = da
        self.db = db
        self.dc = dc
        self.no_yield_surf = no_yield_surf
        self.yield_surf = yield_surf
        self.liquefac1 = liquefac1
        self.liquefac2 = liquefac2
        self.pa = pa
        self.s0 = s0
        if osi is not None:
            osi.n_mat += 1
            self._tag = osi.n_mat
        self._parameters = [self.op_type, self._tag, self.nd, self.rho, self.ref_shear_modul, self.ref_bulk_modul, self.friction_ang, self.peak_shear_stra, self.ref_press, self.press_depend_coe, self.pt_ang, self.ca, self.cb, self.cc, self.cd, self.ce, self.da, self.db, self.dc, self.no_yield_surf]
        special_pms = ['yield_surf', 'liquefac1', 'liquefac2', 'pa', 's0']
        packets = [True, False, False, False, False]
        for i, pm in enumerate(special_pms):
            if getattr(self, pm) is not None:
                if packets[i]:
                    self._parameters += [*getattr(self, pm)]
                else:
                    self._parameters += [getattr(self, pm)]
            else:
                break
        if osi is None:
            self.built = 0
        if osi is not None:
            self.to_process(osi)

    def set_update_material_stage(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'updateMaterialStage', value, ele, eles)

    def set_g_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'shearModulus', value, ele, eles)

    def set_bulk_mod(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'bulkModulus', value, ele, eles)

    def set_friction_angle(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'updateMaterialStage', value, ele, eles)

    def set_cohesion(self, value, ele=None, eles=None):
        self.set_parameter(self.osi, 'shearModulus', value, ele, eles)
